"""
poynting_vector_surface
=======================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class poynting_vector_surface(Operator):
    """Compute the Poynting Vector surface integral

    Parameters
    ----------
    fields_containerA : FieldsContainer
    fields_containerB : FieldsContainer
    fields_containerC : FieldsContainer
    fields_containerD : FieldsContainer
    abstract_meshed_region : MeshedRegion, optional
        The mesh region in this pin have to be
        boundary or skin mesh
    int32 : int, optional
        Load step number, if it's specified, the
        poynting vector is computed only on
        the substeps of this step


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.poynting_vector_surface()

    >>> # Make input connections
    >>> my_fields_containerA = dpf.FieldsContainer()
    >>> op.inputs.fields_containerA.connect(my_fields_containerA)
    >>> my_fields_containerB = dpf.FieldsContainer()
    >>> op.inputs.fields_containerB.connect(my_fields_containerB)
    >>> my_fields_containerC = dpf.FieldsContainer()
    >>> op.inputs.fields_containerC.connect(my_fields_containerC)
    >>> my_fields_containerD = dpf.FieldsContainer()
    >>> op.inputs.fields_containerD.connect(my_fields_containerD)
    >>> my_abstract_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.abstract_meshed_region.connect(my_abstract_meshed_region)
    >>> my_int32 = int()
    >>> op.inputs.int32.connect(my_int32)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.poynting_vector_surface(
    ...     fields_containerA=my_fields_containerA,
    ...     fields_containerB=my_fields_containerB,
    ...     fields_containerC=my_fields_containerC,
    ...     fields_containerD=my_fields_containerD,
    ...     abstract_meshed_region=my_abstract_meshed_region,
    ...     int32=my_int32,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        fields_containerA=None,
        fields_containerB=None,
        fields_containerC=None,
        fields_containerD=None,
        abstract_meshed_region=None,
        int32=None,
        config=None,
        server=None,
    ):
        super().__init__(name="PoyntingVectorSurface", config=config, server=server)
        self._inputs = InputsPoyntingVectorSurface(self)
        self._outputs = OutputsPoyntingVectorSurface(self)
        if fields_containerA is not None:
            self.inputs.fields_containerA.connect(fields_containerA)
        if fields_containerB is not None:
            self.inputs.fields_containerB.connect(fields_containerB)
        if fields_containerC is not None:
            self.inputs.fields_containerC.connect(fields_containerC)
        if fields_containerD is not None:
            self.inputs.fields_containerD.connect(fields_containerD)
        if abstract_meshed_region is not None:
            self.inputs.abstract_meshed_region.connect(abstract_meshed_region)
        if int32 is not None:
            self.inputs.int32.connect(int32)

    @staticmethod
    def _spec():
        description = """Compute the Poynting Vector surface integral"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_containerA",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="fields_containerB",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="fields_containerC",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                3: PinSpecification(
                    name="fields_containerD",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                4: PinSpecification(
                    name="abstract_meshed_region",
                    type_names=["abstract_meshed_region"],
                    optional=True,
                    document="""The mesh region in this pin have to be
        boundary or skin mesh""",
                ),
                5: PinSpecification(
                    name="int32",
                    type_names=["int32"],
                    optional=True,
                    document="""Load step number, if it's specified, the
        poynting vector is computed only on
        the substeps of this step""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="PoyntingVectorSurface", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPoyntingVectorSurface
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsPoyntingVectorSurface
        """
        return super().outputs


class InputsPoyntingVectorSurface(_Inputs):
    """Intermediate class used to connect user inputs to
    poynting_vector_surface operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.poynting_vector_surface()
    >>> my_fields_containerA = dpf.FieldsContainer()
    >>> op.inputs.fields_containerA.connect(my_fields_containerA)
    >>> my_fields_containerB = dpf.FieldsContainer()
    >>> op.inputs.fields_containerB.connect(my_fields_containerB)
    >>> my_fields_containerC = dpf.FieldsContainer()
    >>> op.inputs.fields_containerC.connect(my_fields_containerC)
    >>> my_fields_containerD = dpf.FieldsContainer()
    >>> op.inputs.fields_containerD.connect(my_fields_containerD)
    >>> my_abstract_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.abstract_meshed_region.connect(my_abstract_meshed_region)
    >>> my_int32 = int()
    >>> op.inputs.int32.connect(my_int32)
    """

    def __init__(self, op: Operator):
        super().__init__(poynting_vector_surface._spec().inputs, op)
        self._fields_containerA = Input(
            poynting_vector_surface._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._fields_containerA)
        self._fields_containerB = Input(
            poynting_vector_surface._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._fields_containerB)
        self._fields_containerC = Input(
            poynting_vector_surface._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._fields_containerC)
        self._fields_containerD = Input(
            poynting_vector_surface._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._fields_containerD)
        self._abstract_meshed_region = Input(
            poynting_vector_surface._spec().input_pin(4), 4, op, -1
        )
        self._inputs.append(self._abstract_meshed_region)
        self._int32 = Input(poynting_vector_surface._spec().input_pin(5), 5, op, -1)
        self._inputs.append(self._int32)

    @property
    def fields_containerA(self):
        """Allows to connect fields_containerA input to the operator.

        Parameters
        ----------
        my_fields_containerA : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> op.inputs.fields_containerA.connect(my_fields_containerA)
        >>> # or
        >>> op.inputs.fields_containerA(my_fields_containerA)
        """
        return self._fields_containerA

    @property
    def fields_containerB(self):
        """Allows to connect fields_containerB input to the operator.

        Parameters
        ----------
        my_fields_containerB : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> op.inputs.fields_containerB.connect(my_fields_containerB)
        >>> # or
        >>> op.inputs.fields_containerB(my_fields_containerB)
        """
        return self._fields_containerB

    @property
    def fields_containerC(self):
        """Allows to connect fields_containerC input to the operator.

        Parameters
        ----------
        my_fields_containerC : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> op.inputs.fields_containerC.connect(my_fields_containerC)
        >>> # or
        >>> op.inputs.fields_containerC(my_fields_containerC)
        """
        return self._fields_containerC

    @property
    def fields_containerD(self):
        """Allows to connect fields_containerD input to the operator.

        Parameters
        ----------
        my_fields_containerD : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> op.inputs.fields_containerD.connect(my_fields_containerD)
        >>> # or
        >>> op.inputs.fields_containerD(my_fields_containerD)
        """
        return self._fields_containerD

    @property
    def abstract_meshed_region(self):
        """Allows to connect abstract_meshed_region input to the operator.

        The mesh region in this pin have to be
        boundary or skin mesh

        Parameters
        ----------
        my_abstract_meshed_region : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> op.inputs.abstract_meshed_region.connect(my_abstract_meshed_region)
        >>> # or
        >>> op.inputs.abstract_meshed_region(my_abstract_meshed_region)
        """
        return self._abstract_meshed_region

    @property
    def int32(self):
        """Allows to connect int32 input to the operator.

        Load step number, if it's specified, the
        poynting vector is computed only on
        the substeps of this step

        Parameters
        ----------
        my_int32 : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> op.inputs.int32.connect(my_int32)
        >>> # or
        >>> op.inputs.int32(my_int32)
        """
        return self._int32


class OutputsPoyntingVectorSurface(_Outputs):
    """Intermediate class used to get outputs from
    poynting_vector_surface operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.poynting_vector_surface()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(poynting_vector_surface._spec().outputs, op)
        self._fields_container = Output(
            poynting_vector_surface._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.poynting_vector_surface()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container



"""
contact_total_stress
====================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class contact_total_stress(Operator):
    """Read/compute element contact total stress (pressure plus friction) by
    calling the readers defined by the datasources.

    Parameters
    ----------
    time_scoping : Scoping or int or float or Field, optional
        Time/freq values (use doubles or field),
        time/freq set ids (use ints or
        scoping) or time/freq step ids (use
        scoping with timefreq_steps location)
        required in output.to specify
        time/freq values at specific load
        steps, put a field (and not a list)
        in input with a scoping located on
        "timefreq_steps".linear time freq
        intrapolation is performed if the
        values are not in the result files
        and the data at the max time or freq
        is taken when time/freqs are higher
        than available time/freqs in result
        files.
    mesh_scoping : ScopingsContainer or Scoping, optional
        Nodes or elements scoping required in output.
        the output fields will be scoped on
        these node or element ids. to figure
        out the ordering of the fields data,
        look at their scoping ids as they
        might not be ordered as the input
        scoping was. the scoping's location
        indicates whether nodes or elements
        are asked for. using scopings
        container allows you to split the
        result fields container into domains
    fields_container : FieldsContainer, optional
        Fields container already allocated modified
        inplace
    streams_container : StreamsContainer, optional
        Result file container allowed to be kept open
        to cache data
    data_sources : DataSources
        Result file path container, used if no
        streams are set
    bool_rotate_to_global : bool, optional
        If true the field is rotated to global
        coordinate system (default true)
    mesh : MeshedRegion or MeshesContainer, optional
        Prevents from reading the mesh in the result
        files
    read_cyclic : int, optional
        If 0 cyclic symmetry is ignored, if 1 cyclic
        sector is read, if 2 cyclic expansion
        is done, if 3 cyclic expansion is
        done and stages are merged (default
        is 1)


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.contact_total_stress()

    >>> # Make input connections
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.ScopingsContainer()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_bool_rotate_to_global = bool()
    >>> op.inputs.bool_rotate_to_global.connect(my_bool_rotate_to_global)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_read_cyclic = int()
    >>> op.inputs.read_cyclic.connect(my_read_cyclic)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.contact_total_stress(
    ...     time_scoping=my_time_scoping,
    ...     mesh_scoping=my_mesh_scoping,
    ...     fields_container=my_fields_container,
    ...     streams_container=my_streams_container,
    ...     data_sources=my_data_sources,
    ...     bool_rotate_to_global=my_bool_rotate_to_global,
    ...     mesh=my_mesh,
    ...     read_cyclic=my_read_cyclic,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        time_scoping=None,
        mesh_scoping=None,
        fields_container=None,
        streams_container=None,
        data_sources=None,
        bool_rotate_to_global=None,
        mesh=None,
        read_cyclic=None,
        config=None,
        server=None,
    ):
        super().__init__(name="ECT_STOT", config=config, server=server)
        self._inputs = InputsContactTotalStress(self)
        self._outputs = OutputsContactTotalStress(self)
        if time_scoping is not None:
            self.inputs.time_scoping.connect(time_scoping)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if streams_container is not None:
            self.inputs.streams_container.connect(streams_container)
        if data_sources is not None:
            self.inputs.data_sources.connect(data_sources)
        if bool_rotate_to_global is not None:
            self.inputs.bool_rotate_to_global.connect(bool_rotate_to_global)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)
        if read_cyclic is not None:
            self.inputs.read_cyclic.connect(read_cyclic)

    @staticmethod
    def _spec():
        description = """Read/compute element contact total stress (pressure plus friction) by
            calling the readers defined by the datasources."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_scoping",
                    type_names=[
                        "scoping",
                        "int32",
                        "vector<int32>",
                        "double",
                        "field",
                        "vector<double>",
                    ],
                    optional=True,
                    document="""Time/freq values (use doubles or field),
        time/freq set ids (use ints or
        scoping) or time/freq step ids (use
        scoping with timefreq_steps location)
        required in output.to specify
        time/freq values at specific load
        steps, put a field (and not a list)
        in input with a scoping located on
        "timefreq_steps".linear time freq
        intrapolation is performed if the
        values are not in the result files
        and the data at the max time or freq
        is taken when time/freqs are higher
        than available time/freqs in result
        files.""",
                ),
                1: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scopings_container", "scoping"],
                    optional=True,
                    document="""Nodes or elements scoping required in output.
        the output fields will be scoped on
        these node or element ids. to figure
        out the ordering of the fields data,
        look at their scoping ids as they
        might not be ordered as the input
        scoping was. the scoping's location
        indicates whether nodes or elements
        are asked for. using scopings
        container allows you to split the
        result fields container into domains""",
                ),
                2: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=True,
                    document="""Fields container already allocated modified
        inplace""",
                ),
                3: PinSpecification(
                    name="streams_container",
                    type_names=["streams_container"],
                    optional=True,
                    document="""Result file container allowed to be kept open
        to cache data""",
                ),
                4: PinSpecification(
                    name="data_sources",
                    type_names=["data_sources"],
                    optional=False,
                    document="""Result file path container, used if no
        streams are set""",
                ),
                5: PinSpecification(
                    name="bool_rotate_to_global",
                    type_names=["bool"],
                    optional=True,
                    document="""If true the field is rotated to global
        coordinate system (default true)""",
                ),
                7: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region", "meshes_container"],
                    optional=True,
                    document="""Prevents from reading the mesh in the result
        files""",
                ),
                14: PinSpecification(
                    name="read_cyclic",
                    type_names=["enum dataProcessing::ECyclicReading", "int32"],
                    optional=True,
                    document="""If 0 cyclic symmetry is ignored, if 1 cyclic
        sector is read, if 2 cyclic expansion
        is done, if 3 cyclic expansion is
        done and stages are merged (default
        is 1)""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="ECT_STOT", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsContactTotalStress
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsContactTotalStress
        """
        return super().outputs


class InputsContactTotalStress(_Inputs):
    """Intermediate class used to connect user inputs to
    contact_total_stress operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.contact_total_stress()
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.ScopingsContainer()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_bool_rotate_to_global = bool()
    >>> op.inputs.bool_rotate_to_global.connect(my_bool_rotate_to_global)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_read_cyclic = int()
    >>> op.inputs.read_cyclic.connect(my_read_cyclic)
    """

    def __init__(self, op: Operator):
        super().__init__(contact_total_stress._spec().inputs, op)
        self._time_scoping = Input(contact_total_stress._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._time_scoping)
        self._mesh_scoping = Input(contact_total_stress._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._mesh_scoping)
        self._fields_container = Input(
            contact_total_stress._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._fields_container)
        self._streams_container = Input(
            contact_total_stress._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._streams_container)
        self._data_sources = Input(contact_total_stress._spec().input_pin(4), 4, op, -1)
        self._inputs.append(self._data_sources)
        self._bool_rotate_to_global = Input(
            contact_total_stress._spec().input_pin(5), 5, op, -1
        )
        self._inputs.append(self._bool_rotate_to_global)
        self._mesh = Input(contact_total_stress._spec().input_pin(7), 7, op, -1)
        self._inputs.append(self._mesh)
        self._read_cyclic = Input(
            contact_total_stress._spec().input_pin(14), 14, op, -1
        )
        self._inputs.append(self._read_cyclic)

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator.

        Time/freq values (use doubles or field),
        time/freq set ids (use ints or
        scoping) or time/freq step ids (use
        scoping with timefreq_steps location)
        required in output.to specify
        time/freq values at specific load
        steps, put a field (and not a list)
        in input with a scoping located on
        "timefreq_steps".linear time freq
        intrapolation is performed if the
        values are not in the result files
        and the data at the max time or freq
        is taken when time/freqs are higher
        than available time/freqs in result
        files.

        Parameters
        ----------
        my_time_scoping : Scoping or int or float or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> # or
        >>> op.inputs.time_scoping(my_time_scoping)
        """
        return self._time_scoping

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Nodes or elements scoping required in output.
        the output fields will be scoped on
        these node or element ids. to figure
        out the ordering of the fields data,
        look at their scoping ids as they
        might not be ordered as the input
        scoping was. the scoping's location
        indicates whether nodes or elements
        are asked for. using scopings
        container allows you to split the
        result fields container into domains

        Parameters
        ----------
        my_mesh_scoping : ScopingsContainer or Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Fields container already allocated modified
        inplace

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def streams_container(self):
        """Allows to connect streams_container input to the operator.

        Result file container allowed to be kept open
        to cache data

        Parameters
        ----------
        my_streams_container : StreamsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.streams_container.connect(my_streams_container)
        >>> # or
        >>> op.inputs.streams_container(my_streams_container)
        """
        return self._streams_container

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator.

        Result file path container, used if no
        streams are set

        Parameters
        ----------
        my_data_sources : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> # or
        >>> op.inputs.data_sources(my_data_sources)
        """
        return self._data_sources

    @property
    def bool_rotate_to_global(self):
        """Allows to connect bool_rotate_to_global input to the operator.

        If true the field is rotated to global
        coordinate system (default true)

        Parameters
        ----------
        my_bool_rotate_to_global : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.bool_rotate_to_global.connect(my_bool_rotate_to_global)
        >>> # or
        >>> op.inputs.bool_rotate_to_global(my_bool_rotate_to_global)
        """
        return self._bool_rotate_to_global

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        Prevents from reading the mesh in the result
        files

        Parameters
        ----------
        my_mesh : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh

    @property
    def read_cyclic(self):
        """Allows to connect read_cyclic input to the operator.

        If 0 cyclic symmetry is ignored, if 1 cyclic
        sector is read, if 2 cyclic expansion
        is done, if 3 cyclic expansion is
        done and stages are merged (default
        is 1)

        Parameters
        ----------
        my_read_cyclic : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> op.inputs.read_cyclic.connect(my_read_cyclic)
        >>> # or
        >>> op.inputs.read_cyclic(my_read_cyclic)
        """
        return self._read_cyclic


class OutputsContactTotalStress(_Outputs):
    """Intermediate class used to get outputs from
    contact_total_stress operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.contact_total_stress()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(contact_total_stress._spec().outputs, op)
        self._fields_container = Output(
            contact_total_stress._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.contact_total_stress()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container



"""
phase_of_max
============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class phase_of_max(Operator):
    """Evaluates phase of maximum.

    Parameters
    ----------
    real_field : Field
    imaginary_field : Field
    abs_value : bool, optional
        Should use absolute value.
    phase_increment : float
        Phase increment.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.min_max.phase_of_max()

    >>> # Make input connections
    >>> my_real_field = dpf.Field()
    >>> op.inputs.real_field.connect(my_real_field)
    >>> my_imaginary_field = dpf.Field()
    >>> op.inputs.imaginary_field.connect(my_imaginary_field)
    >>> my_abs_value = bool()
    >>> op.inputs.abs_value.connect(my_abs_value)
    >>> my_phase_increment = float()
    >>> op.inputs.phase_increment.connect(my_phase_increment)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.min_max.phase_of_max(
    ...     real_field=my_real_field,
    ...     imaginary_field=my_imaginary_field,
    ...     abs_value=my_abs_value,
    ...     phase_increment=my_phase_increment,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(
        self,
        real_field=None,
        imaginary_field=None,
        abs_value=None,
        phase_increment=None,
        config=None,
        server=None,
    ):
        super().__init__(name="phase_of_max", config=config, server=server)
        self._inputs = InputsPhaseOfMax(self)
        self._outputs = OutputsPhaseOfMax(self)
        if real_field is not None:
            self.inputs.real_field.connect(real_field)
        if imaginary_field is not None:
            self.inputs.imaginary_field.connect(imaginary_field)
        if abs_value is not None:
            self.inputs.abs_value.connect(abs_value)
        if phase_increment is not None:
            self.inputs.phase_increment.connect(phase_increment)

    @staticmethod
    def _spec():
        description = """Evaluates phase of maximum."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="real_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="imaginary_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="abs_value",
                    type_names=["bool"],
                    optional=True,
                    document="""Should use absolute value.""",
                ),
                3: PinSpecification(
                    name="phase_increment",
                    type_names=["double"],
                    optional=False,
                    document="""Phase increment.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="phase_of_max", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPhaseOfMax
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsPhaseOfMax
        """
        return super().outputs


class InputsPhaseOfMax(_Inputs):
    """Intermediate class used to connect user inputs to
    phase_of_max operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.min_max.phase_of_max()
    >>> my_real_field = dpf.Field()
    >>> op.inputs.real_field.connect(my_real_field)
    >>> my_imaginary_field = dpf.Field()
    >>> op.inputs.imaginary_field.connect(my_imaginary_field)
    >>> my_abs_value = bool()
    >>> op.inputs.abs_value.connect(my_abs_value)
    >>> my_phase_increment = float()
    >>> op.inputs.phase_increment.connect(my_phase_increment)
    """

    def __init__(self, op: Operator):
        super().__init__(phase_of_max._spec().inputs, op)
        self._real_field = Input(phase_of_max._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._real_field)
        self._imaginary_field = Input(phase_of_max._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._imaginary_field)
        self._abs_value = Input(phase_of_max._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._abs_value)
        self._phase_increment = Input(phase_of_max._spec().input_pin(3), 3, op, -1)
        self._inputs.append(self._phase_increment)

    @property
    def real_field(self):
        """Allows to connect real_field input to the operator.

        Parameters
        ----------
        my_real_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.min_max.phase_of_max()
        >>> op.inputs.real_field.connect(my_real_field)
        >>> # or
        >>> op.inputs.real_field(my_real_field)
        """
        return self._real_field

    @property
    def imaginary_field(self):
        """Allows to connect imaginary_field input to the operator.

        Parameters
        ----------
        my_imaginary_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.min_max.phase_of_max()
        >>> op.inputs.imaginary_field.connect(my_imaginary_field)
        >>> # or
        >>> op.inputs.imaginary_field(my_imaginary_field)
        """
        return self._imaginary_field

    @property
    def abs_value(self):
        """Allows to connect abs_value input to the operator.

        Should use absolute value.

        Parameters
        ----------
        my_abs_value : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.min_max.phase_of_max()
        >>> op.inputs.abs_value.connect(my_abs_value)
        >>> # or
        >>> op.inputs.abs_value(my_abs_value)
        """
        return self._abs_value

    @property
    def phase_increment(self):
        """Allows to connect phase_increment input to the operator.

        Phase increment.

        Parameters
        ----------
        my_phase_increment : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.min_max.phase_of_max()
        >>> op.inputs.phase_increment.connect(my_phase_increment)
        >>> # or
        >>> op.inputs.phase_increment(my_phase_increment)
        """
        return self._phase_increment


class OutputsPhaseOfMax(_Outputs):
    """Intermediate class used to get outputs from
    phase_of_max operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.min_max.phase_of_max()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(phase_of_max._spec().outputs, op)
        self._field = Output(phase_of_max._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.min_max.phase_of_max()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



"""
Plotter
=======
This module contains the DPF plotter class.

Contains classes used to plot a mesh and a fields container using PyVista.
"""

from __future__ import annotations

import tempfile
import os
import sys
import numpy as np
import warnings
from typing import TYPE_CHECKING, List, Union

from ansys import dpf
from ansys.dpf import core
from ansys.dpf.core.common import locations, DefinitionLabels
from ansys.dpf.core.common import shell_layers as eshell_layers
from ansys.dpf.core.helpers.streamlines import _sort_supported_kwargs
from ansys.dpf.core import errors as dpf_errors
from ansys.dpf.core.nodes import Node, Nodes

if TYPE_CHECKING:  # pragma: no cover
    from ansys.dpf.core.meshed_region import MeshedRegion


class _InternalPlotterFactory:
    """
    Factory for _InternalPlotter based on the backend."""

    @staticmethod
    def get_plotter_class():
        return _PyVistaPlotter


class _PyVistaPlotter:
    """The _InternalPlotter class is based on PyVista."""

    def __init__(self, **kwargs):
        # Import pyvista
        from ansys.dpf.core.vtk_helper import PyVistaImportError

        try:
            import pyvista as pv
        except ModuleNotFoundError:
            raise PyVistaImportError

        # Filter kwargs
        kwargs_in = _sort_supported_kwargs(bound_method=pv.Plotter.__init__, **kwargs)
        # Initiate pyvista Plotter
        self._plotter = pv.Plotter(**kwargs_in)

    def add_scale_factor_legend(self, scale_factor, **kwargs):
        kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.add_text, **kwargs)
        _ = kwargs_in.pop("position", None)
        _ = kwargs_in.pop("font_size", None)
        _ = kwargs_in.pop("text", None)
        _ = kwargs_in.pop("color", None)
        self._plotter.add_text(
            f"Scale factor: {scale_factor}",
            position="upper_right",
            font_size=12,
            **kwargs_in,
        )

    def add_points(self, points, field, **kwargs):
        import pyvista as pv

        point_cloud = pv.PolyData(points)
        if field:
            point_cloud[f"{field.name}"] = field.data
        self._plotter.add_points(point_cloud, **kwargs)

    def add_line(self, points, field=None, **kwargs):
        import pyvista as pv

        line_field = pv.PolyData(np.array(points))
        if field:
            line_field[f"{field.name}"] = field.data
            self._plotter.add_mesh(line_field, **kwargs)
        else:
            self._plotter.add_lines(points, **kwargs)

    def add_plane(self, plane, field=None, **kwargs):
        import pyvista as pv

        plane_plot = pv.Plane(
            center=plane.center,
            direction=plane.normal_dir,
            i_size=plane.width,
            j_size=plane.height,
            i_resolution=plane.n_cells_x,
            j_resolution=plane.n_cells_y,
        )
        if field:
            plane[f"{field.name}"] = field.data
        self._plotter.add_mesh(plane_plot, **kwargs)

    def add_mesh(self, meshed_region, deform_by=None, scale_factor=1.0, as_linear=True, **kwargs):

        kwargs = self._set_scalar_bar_title(kwargs)

        # Set defaults for PyDPF
        kwargs.setdefault("show_edges", True)
        kwargs.setdefault("nan_color", "grey")

        # If deformed geometry, print the scale_factor
        if deform_by:
            self.add_scale_factor_legend(scale_factor, **kwargs)

        # Filter kwargs
        kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.add_mesh, **kwargs)
        # Give the mesh to the pyvista Plotter
        # Have to remove any active scalar field from the pre-existing grid object,
        # otherwise we get two scalar bars when calling several plot_contour on the same mesh
        # but not for the same field. The PyVista UnstructuredGrid keeps memory of it.
        if not deform_by:
            if as_linear != meshed_region.as_linear:
                grid = meshed_region._as_vtk(
                    meshed_region.nodes.coordinates_field, as_linear=as_linear
                )
                meshed_region.as_linear = as_linear
            else:
                grid = meshed_region.grid
        else:
            grid = meshed_region._as_vtk(
                meshed_region.deform_by(deform_by, scale_factor), as_linear=as_linear
            )

        # show axes
        show_axes = kwargs.pop("show_axes", None)
        if show_axes:
            self._plotter.add_axes()

        grid.set_active_scalars(None)
        self._plotter.add_mesh(grid, **kwargs_in)

    def add_point_labels(
        self,
        nodes: Union[Nodes, List[Node], List[int]],
        meshed_region: MeshedRegion,
        labels: Union[List[str], None] = None,
        **kwargs,
    ) -> List:
        label_actors = []
        if isinstance(nodes, Nodes):
            nodes = nodes.scoping.ids
        elif isinstance(nodes, list):
            if isinstance(nodes[0], Node):
                nodes = [node.id for node in nodes]
        node_indexes = [meshed_region.nodes.mapping_id_to_index.get(node_id) for node_id in nodes]
        grid_points = [meshed_region.grid.points[node_index] for node_index in node_indexes]

        def get_label_at_grid_point(index):
            try:
                label = labels[index]
            except:
                label = None
            return label

        # Filter kwargs
        kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.add_point_labels, **kwargs)
        import pyvista as pv

        # The scalar data used will be the one of the last field added.
        from packaging.version import parse

        active_scalars = None
        if parse(pv.__version__) >= parse("0.35.2"):
            for data_set in self._plotter._datasets:
                if type(data_set) is pv.core.pointset.UnstructuredGrid:
                    active_scalars = data_set.active_scalars
        else:
            active_scalars = meshed_region.grid.active_scalars
        if active_scalars is None:
            self.add_mesh(meshed_region=meshed_region)
        # For all grid_points given
        for index, grid_point in enumerate(grid_points):
            # Check for existing label at that point
            label_at_grid_point = get_label_at_grid_point(index)
            if label_at_grid_point:
                # If there is already a label, create the associated actor
                label_actors.append(
                    self._plotter.add_point_labels(grid_point, [labels[index]], **kwargs_in)
                )
            else:
                if active_scalars is not None:
                    # get the value of the current scalar field if present
                    scalar_at_index = active_scalars[node_indexes[index]]
                    value = f"{scalar_at_index:.2f}"
                else:
                    # if no scalar field is present, print the node id
                    value = nodes[index]
                label_actors.append(
                    self._plotter.add_point_labels(grid_point, [value], **kwargs_in)
                )
        return label_actors

    def add_field(
        self,
        field,
        meshed_region=None,
        show_max=False,
        show_min=False,
        label_text_size=30,
        label_point_size=20,
        deform_by=None,
        scale_factor=1.0,
        scale_factor_legend=None,
        as_linear=True,
        **kwargs,
    ):
        # Get the field name
        name = field.name.split("_")[0]
        unit = field.unit
        kwargs.setdefault("stitle", f"{name} ({unit})")

        kwargs = self._set_scalar_bar_title(kwargs)

        kwargs.setdefault("show_edges", True)
        kwargs.setdefault("nan_color", "grey")

        # show axes
        show_axes = kwargs.pop("show_axes", None)
        if show_axes:
            self._plotter.add_axes()

        # get the meshed region location
        if meshed_region is None:
            meshed_region = field.meshed_region

        location = field.location
        if location == locations.nodal:
            mesh_location = meshed_region.nodes
        elif location == locations.elemental:
            mesh_location = meshed_region.elements
            if show_max or show_min:
                warnings.warn("`show_max` and `show_min` is only supported for Nodal results.")
                show_max = False
                show_min = False
        elif location == locations.faces:
            mesh_location = meshed_region.faces
            if len(mesh_location) == 0:
                raise ValueError("No faces found to plot on")
            if show_max or show_min:
                warnings.warn("`show_max` and `show_min` is only supported for Nodal results.")
                show_max = False
                show_min = False
        elif location == locations.overall:
            mesh_location = meshed_region.elements
        else:
            raise ValueError("Only elemental, nodal or faces location are supported for plotting.")
        component_count = field.component_count
        if component_count > 1:
            overall_data = np.full((len(mesh_location), component_count), np.nan)
        else:
            overall_data = np.full(len(mesh_location), np.nan)
        if location != locations.overall:
            ind, mask = mesh_location.map_scoping(field.scoping)
            overall_data[ind] = field.data[mask]
        else:
            overall_data[:] = field.data[0]
        # Filter kwargs for add_mesh
        kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.add_mesh, **kwargs)
        # Have to remove any active scalar field from the pre-existing grid object,
        # otherwise we get two scalar bars when calling several plot_contour on the same mesh
        # but not for the same field. The PyVista UnstructuredGrid keeps memory of it.
        if not deform_by:
            grid = meshed_region.grid
        else:
            grid = meshed_region._as_vtk(
                meshed_region.deform_by(deform_by, scale_factor), as_linear
            )
        grid.set_active_scalars(None)
        self._plotter.add_mesh(grid, scalars=overall_data, **kwargs_in)

        # If deformed geometry, print the scale_factor
        if deform_by and scale_factor_legend is not False:
            if scale_factor_legend is None:
                scale_factor_legend = scale_factor
            self.add_scale_factor_legend(scale_factor_legend, **kwargs)

        if show_max or show_min:
            # Get Min-Max for the field
            min_max = core.operators.min_max.min_max()
            min_max.inputs.connect(field)

        # Add Min and Max Labels
        labels = []
        grid_points = []
        if show_max:
            max_field = min_max.outputs.field_max()
            # Get Node ID at max.
            node_id_at_max = max_field.scoping.id(0)
            labels.append(f"Max: {max_field.data[0]:.2f}\nNodeID: {node_id_at_max}")
            # Get Node index at max value.
            node_index_at_max = meshed_region.nodes.scoping.index(node_id_at_max)
            # Append the corresponding Grid Point.
            grid_points.append(meshed_region.grid.points[node_index_at_max])

        if show_min:
            min_field = min_max.outputs.field_min()
            # Get Node ID at min.
            node_id_at_min = min_field.scoping.id(0)
            labels.append(f"Min: {min_field.data[0]:.2f}\nNodeID: {node_id_at_min}")
            # Get Node index at min. value.
            node_index_at_min = meshed_region.nodes.scoping.index(node_id_at_min)
            # Append the corresponding Grid Point.
            grid_points.append(meshed_region.grid.points[node_index_at_min])

        # Plot labels:
        for index, grid_point in enumerate(grid_points):
            self._plotter.add_point_labels(
                grid_point,
                [labels[index]],
                font_size=label_text_size,
                point_size=label_point_size,
            )

    def add_streamlines(self, streamlines, source=None, radius=1.0, **kwargs):
        permissive = kwargs.pop("permissive", True)
        kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.add_mesh, **kwargs)
        # set streamline on plotter
        sargs = dict(vertical=False)
        streamlines = streamlines._as_pyvista_data_set()
        if not (permissive and streamlines.n_points == 0):
            self._plotter.add_mesh(
                streamlines.tube(radius=radius), scalar_bar_args=sargs, **kwargs_in
            )
        if source is not None:
            src = source._as_pyvista_data_set()
            self._plotter.add_mesh(src, **kwargs_in)

    def show_figure(self, **kwargs):

        text = kwargs.pop("text", None)
        if text is not None:
            self._plotter.add_text(text, position="lower_edge")

        background = kwargs.pop("background", None)
        if background is not None:
            self._plotter.set_background(background)

        # show result
        show_axes = kwargs.pop("show_axes", None)
        if show_axes:
            self._plotter.add_axes()

        # Set cpos
        cpos = kwargs.pop("cpos", None)
        if cpos is not None:
            self._plotter.camera_position = cpos

        # Show depending on return_cpos option
        kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.show, **kwargs)
        return self._plotter.show(**kwargs_in)

    @staticmethod
    def _set_scalar_bar_title(kwargs):
        stitle = kwargs.pop("stitle", None)
        # use scalar_bar_args
        scalar_bar_args = kwargs.pop("scalar_bar_args", None)
        if not scalar_bar_args:
            scalar_bar_args = {"title": stitle}
        kwargs.setdefault("scalar_bar_args", scalar_bar_args)
        return kwargs


class DpfPlotter:
    """DpfPlotter class. Can be used in order to plot
    results over a mesh.

    The current DpfPlotter is a PyVista based object.

    That means that PyVista must be installed, and that
    it supports kwargs as parameter (the argument
    must be supported by the installed PyVista version).
    More information about the available arguments are
    available at :class:`pyvista.Plotter`.
    """

    def __init__(self, **kwargs):
        """Create a DpfPlotter object.

        The current DpfPlotter is a PyVista based object.

        That means that PyVista must be installed, and that
        it supports **kwargs as parameter (the argument
        must be supported by the installed PyVista version).
        More information about the available arguments are
        available at :class:`pyvista.Plotter`.

        Parameters
        ----------
        **kwargs : optional
            Additional keyword arguments for the plotter. More information
            are available at :class:`pyvista.Plotter`.

        Examples
        --------
        >>> from ansys.dpf.core.plotter import DpfPlotter
        >>> pl = DpfPlotter(notebook=False)

        """
        _InternalPlotterClass = _InternalPlotterFactory.get_plotter_class()
        self._internal_plotter = _InternalPlotterClass(**kwargs)
        self._labels = []

    @property
    def labels(self):
        """Return a list of labels.

        Returns
        --------
        list
            List of Label(s). Each list member or member group
            will share same properties.
        """
        return self._labels

    def add_node_labels(
        self,
        nodes: Union[Nodes, List[Node], List[int]],
        meshed_region: MeshedRegion,
        labels: Union[List[str], None] = None,
        **kwargs,
    ):
        """Add labels at nodal locations.

        Parameters
        ----------
        nodes :
            Nodes where the labels should be added.
        meshed_region:
            MeshedRegion to plot.
        labels:
            The labels to use. A node for which the label is not defined or `None`
            will show the scalar value of the currently active field at that node,
            or, if no field is active, its node ID.
        kwargs:
            Keyword arguments controlling label properties.
            See :func:`pyvista.Plotter.add_point_labels`.
        """
        self._labels.append(
            self._internal_plotter.add_point_labels(
                nodes=nodes, meshed_region=meshed_region, labels=labels, **kwargs
            )
        )

    def add_points(self, points, field=None, **kwargs):
        self._internal_plotter.add_points(points, field, **kwargs)

    def add_line(self, points, field=None, **kwargs):
        self._internal_plotter.add_line(points, field, **kwargs)

    def add_plane(self, plane, field=None, **kwargs):
        self._internal_plotter.add_plane(plane, field, **kwargs)

    def add_mesh(self, meshed_region, deform_by=None, scale_factor=1.0, **kwargs):
        """Add a mesh to plot.

        Parameters
        ----------
        meshed_region : MeshedRegion
            MeshedRegion to plot.
        deform_by : Field, Result, Operator, optional
            Used to deform the plotted mesh. Must output a 3D vector field.
            Defaults to None.
        scale_factor : float, optional
            Scaling factor to apply when warping the mesh. Defaults to 1.0.
        **kwargs : optional
            Additional keyword arguments for the plotter. More information
            are available at :func:`pyvista.plot`.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_multishells_rst())
        >>> mesh = model.metadata.meshed_region
        >>> from ansys.dpf.core.plotter import DpfPlotter
        >>> pl = DpfPlotter()
        >>> pl.add_mesh(mesh)

        """
        if meshed_region.grid is not None:
            meshed_region.grid.clear_data()
        self._internal_plotter.add_mesh(
            meshed_region=meshed_region,
            deform_by=deform_by,
            scale_factor=scale_factor,
            as_linear=True,
            **kwargs,
        )

    def add_streamlines(
        self,
        streamlines,
        source=None,
        radius=0.1,
        **kwargs,
    ):
        """Add a streamline to the plotter.

        The current add_streamlines method adds streamline
        as a PyVista based object.
        For more information about arguments, see
        :func:`pyvista.DataSetFilters.streamlines`.

        Parameters
        ----------
        streamlines : helpers.streamlines.Streamlines
            Object containing computed streamlines data,
            computed using `helpers.streamlines.compute_streamlines`
            function.
        source : helpers.streamlines.StreamlinesSource, optional
            Object containing computed streamines source data,
            computed using `helpers.streamlines.compute_streamlines`
            function.
        **kwargs : optional
            Additional keyword arguments for the plotter. More information
            is available at :func:`pyvista.plot`.
            The "permissive" (boolean, default being True) can be used to
            avoid throwing if computed streamlines are empty. See
            ``Examples`` section for more information.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> from ansys.dpf.core.helpers.streamlines import compute_streamlines
        >>> # Get model and meshed region
        >>> files = examples.download_fluent_mixing_elbow_steady_state()
        >>> ds = dpf.DataSources()
        >>> ds.set_result_file_path(files["cas"][0], "cas")
        >>> ds.add_file_path(files["dat"][1], "dat")
        >>> model = dpf.Model(ds)
        >>> mesh = model.metadata.meshed_region
        >>> # Get velocity data
        >>> velocity_op = model.results.velocity()
        >>> fc = velocity_op.outputs.fields_container()
        >>> op = dpf.operators.averaging.to_nodal_fc(fields_container=fc)
        >>> field = op.outputs.fields_container()[0]
        >>> # Plot
        >>> from ansys.dpf.core.plotter import DpfPlotter
        >>> pl = DpfPlotter()
        >>> pl.add_mesh(meshed_region=mesh, opacity=0.15, color="g")
        >>> streamline_obj = compute_streamlines(
        ...        meshed_region=mesh,
        ...        field=field,
        ...        source_center=(0.55, 0.55, 0.),
        ...        n_points=10,
        ...        source_radius=0.08,
        ...        max_time=10.0
        ...        )
        >>> pl.add_streamlines(
        ...        streamlines=streamline_obj,
        ...        radius=0.001,
        ...        )
        >>> pl.show_figure(show_axes=True)

        """
        self._internal_plotter.add_streamlines(
            streamlines=streamlines,
            source=source,
            radius=radius,
            **kwargs,
        )

    def add_field(
        self,
        field,
        meshed_region=None,
        show_max=False,
        show_min=False,
        label_text_size=30,
        label_point_size=20,
        deform_by=None,
        scale_factor=1.0,
        **kwargs,
    ):
        """Add a field containing data to the plotter.

        A meshed_region to plot on can be added.
        If no ``meshed_region`` is specified, the field
        support will be used. Ensure that the field
        support is a ``meshed_region``.

        Parameters
        ----------
        field : Field
            Field data to plot
        meshed_region : MeshedRegion, optional
            ``MeshedRegion`` to plot the field on.
        show_max : bool, optional
            Label the point with the maximum value.
        show_min : bool, optional
            Label the point with the minimum value.
        deform_by : Field, Result, Operator, optional
            Used to deform the plotted mesh. Must output a 3D vector field.
            Defaults to None.
        scale_factor : float, optional
            Scaling factor to apply when warping the mesh. Defaults to 1.0.
        **kwargs : optional
            Additional keyword arguments for the plotter. More information
            are available at :func:`pyvista.plot`.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_multishells_rst())
        >>> mesh = model.metadata.meshed_region
        >>> field = model.results.displacement().outputs.fields_container()[0]
        >>> from ansys.dpf.core.plotter import DpfPlotter
        >>> pl = DpfPlotter()
        >>> pl.add_field(field, mesh)

        """
        self._internal_plotter.add_field(
            field=field,
            meshed_region=meshed_region,
            show_max=show_max,
            show_min=show_min,
            label_text_size=label_text_size,
            label_point_size=label_point_size,
            deform_by=deform_by,
            scale_factor=scale_factor,
            as_linear=True,
            **kwargs,
        )

    def show_figure(self, **kwargs):
        """Plot the figure built by the plotter object.

        Parameters
        ----------
        **kwargs : optional
            Additional keyword arguments for the plotter. More information
            are available at :func:`pyvista.plot`.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_multishells_rst())
        >>> mesh = model.metadata.meshed_region
        >>> field = model.results.displacement().outputs.fields_container()[0]
        >>> from ansys.dpf.core.plotter import DpfPlotter
        >>> pl = DpfPlotter()
        >>> pl.add_field(field, mesh)
        >>> pl.show_figure()

        """
        if "notebook" in kwargs.keys():
            warnings.simplefilter("once")
            warnings.warn(
                "'notebook' is not a valid kwarg for show_figure(). "
                "Please give this argument to the init of DpfPlotter."
            )
        return self._internal_plotter.show_figure(**kwargs)


def plot_chart(fields_container, off_screen=False, screenshot=None):
    """Plot the minimum/maximum result values over time.

    This is a valid method if ``time_freq_support`` contains
    several time_steps, such as in a transient analysis.

    Parameters
    ----------
    fields_container : dpf.core.FieldsContainer
        Fields container that must contains a result for each
        time step of ``time_freq_support``.
    off_screen : bool, optional
        Whether to render the image off-screen. Useful for batch workflows.
        The default is ``False``.
    screenshot : path-like, optional
        A file path to which the figure should be saved. The format is inferred from the file
        extension in the path (defaults to ".png"). The default is ``None``.


    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.find_transient_therm())
    >>> t = model.results.temperature.on_all_time_freqs()
    >>> fc = t.outputs.fields_container()
    >>> plotter = dpf.plotter.plot_chart(fc)

    """
    p = Plotter(None)
    return p.plot_chart(fields_container, screenshot=screenshot, off_screen=off_screen)


class Plotter:
    """Plots fields and meshed regions in DPF-Core.

    Parameters
    ----------
    mesh : str
        Name of the mesh.

    """

    def __init__(self, mesh, **kwargs):
        _InternalPlotterClass = _InternalPlotterFactory.get_plotter_class()
        self._internal_plotter = _InternalPlotterClass(mesh=mesh, **kwargs)
        self._mesh = mesh

    def plot_mesh(self, **kwargs):
        """Plot the mesh using PyVista.

        Parameters
        ----------
        notebook : bool, optional
            When ``None`` (default) plot a static image within an
            iPython notebook if available.  When ``False``, plot
            external to the notebook with an interactive window.  When
            ``True``, always plot within a notebook.
        **kwargs : optional
            Additional keyword arguments for the plotter. For more information,
            ee ``help(pyvista.plot)``.

        """
        kwargs.setdefault("color", "w")
        kwargs.setdefault("show_edges", True)
        return self._mesh.grid.plot(**kwargs)

    @staticmethod
    def plot_chart(fields_container, off_screen=False, screenshot=None):
        """Plot the minimum/maximum result values over time.

        This is a valid method if ``time_freq_support`` contains
        several time steps, such as in a transient analysis.

        Parameters
        ----------
        fields_container : dpf.core.FieldsContainer
            Fields container that must contain a result for each
            time step of ``time_freq_support``.
        off_screen : bool, optional
            Used to prevent the figure from showing in a pop-up, useful for batch image generation.
            Defaults to False.
        screenshot : str, os.pathLike, optional
            Path to save the figure to. Defaults to None. If no extension is given, defaults to
            .png format. See ``help(matplotlib.pyplot.savefig)`` for more information on
            supported formats.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_simple_bar())
        >>> disp = model.results.displacement()
        >>> scoping = dpf.Scoping()
        >>> scoping.ids = range(1, len(model.metadata.time_freq_support.time_frequencies) + 1)
        >>> disp.inputs.time_scoping.connect(scoping)
        >>> fc = disp.outputs.fields_container()
        >>> plotter = dpf.plotter.Plotter(model.metadata.meshed_region)
        >>> pl = plotter.plot_chart(fc)

        """
        # Import matplotlib.pyplot
        try:
            import matplotlib.pyplot as pyplot
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To use plot_chart capabilities, please install "
                "matplotlib with :\n pip install matplotlib>=3.2"
            )
        tfq = fields_container.time_freq_support
        if len(fields_container) != len(tfq.time_frequencies):
            raise Exception(
                "Fields container must contain real fields at all time "
                "steps of the time_freq_support."
            )
        time_field = tfq.time_frequencies
        normOp = dpf.core.Operator("norm_fc")
        minmaxOp = dpf.core.Operator("min_max_fc")
        normOp.inputs.fields_container.connect(fields_container)
        minmaxOp.inputs.connect(normOp.outputs)
        fieldMin = minmaxOp.outputs.field_min()
        fieldMax = minmaxOp.outputs.field_max()
        pyplot.plot(time_field.data, fieldMax.data, "r", label="Maximum")
        pyplot.plot(time_field.data, fieldMin.data, "b", label="Minimum")
        unit = tfq.time_frequencies.unit
        if unit == "Hz":
            pyplot.xlabel("frequencies (Hz)")
        elif unit == "s":
            pyplot.xlabel("time (s)")
        elif unit is not None:
            pyplot.xlabel(unit)
        substr = fields_container[0].name.split("_")
        pyplot.ylabel(substr[0] + fieldMin.unit)
        pyplot.title(substr[0] + ": min/max values over time")
        pyplot.legend()
        f = pyplot.gcf()
        if screenshot:
            f.savefig(screenshot)
        if not off_screen:
            pyplot.show(block=True)
        return f

    def plot_contour(
        self,
        field_or_fields_container,
        shell_layers=None,
        meshed_region=None,
        deform_by=None,
        scale_factor=1.0,
        **kwargs,
    ):
        """Plot the contour result on its mesh support.

        You cannot plot a fields container containing results at several
        time steps.

        Parameters
        ----------
        field_or_fields_container : dpf.core.Field or dpf.core.FieldsContainer
            Field or field container that contains the result to plot.
        shell_layers : core.shell_layers, optional
            Enum used to set the shell layers if the model to plot
            contains shell elements.
        deform_by : Field, Result, Operator, optional
            Used to deform the plotted mesh. Must output a 3D vector field.
            Defaults to None.
        scale_factor : float, optional
            Scaling factor to apply when warping the mesh. Defaults to 1.0.
        **kwargs : optional
            Additional keyword arguments for the plotter. For more information,
            see ``help(pyvista.plot)``.
        """
        if not sys.warnoptions:
            import warnings

            warnings.simplefilter("ignore")

        if isinstance(field_or_fields_container, (dpf.core.Field, dpf.core.FieldsContainer)):
            fields_container = None
            if isinstance(field_or_fields_container, dpf.core.Field):
                fields_container = dpf.core.FieldsContainer(
                    server=field_or_fields_container._server
                )
                fields_container.add_label(DefinitionLabels.time)
                fields_container.add_field({DefinitionLabels.time: 1}, field_or_fields_container)
            elif isinstance(field_or_fields_container, dpf.core.FieldsContainer):
                fields_container = field_or_fields_container
        else:
            raise TypeError("Only field or fields_container can be plotted.")

        # pre-loop to check if the there are several time steps
        labels = fields_container.get_label_space(0)
        if DefinitionLabels.complex in labels.keys():
            raise dpf_errors.ComplexPlottingError
        if DefinitionLabels.time in labels.keys():
            first_time = labels[DefinitionLabels.time]
            for i in range(1, len(fields_container)):
                label = fields_container.get_label_space(i)
                if label[DefinitionLabels.time] != first_time:
                    raise dpf_errors.FieldContainerPlottingError

        if meshed_region is not None:
            mesh = meshed_region
        else:
            mesh = self._mesh

        # get mesh scoping
        location = None
        component_count = None
        name = None

        # pre-loop to get location and component count
        for field in fields_container:
            if len(field.data) != 0:
                location = field.location
                component_count = field.component_count
                name = field.name.split("_")[0]
                unit = field.unit
                break

        if location == locations.nodal:
            mesh_location = mesh.nodes
        elif location == locations.elemental:
            mesh_location = mesh.elements
        elif location == locations.faces:
            mesh_location = mesh.faces
        else:
            raise ValueError("Only elemental, nodal or faces location are supported for plotting.")

        # pre-loop: check if shell layers for each field, if yes, set the shell layers
        changeOp = core.Operator("change_shellLayers")
        for field in fields_container:
            shell_layer_check = field.shell_layers
            if shell_layer_check in [
                eshell_layers.topbottom,
                eshell_layers.topbottommid,
            ]:
                changeOp.inputs.fields_container.connect(fields_container)
                sl = eshell_layers.top
                if shell_layers is not None:
                    if not isinstance(shell_layers, eshell_layers):
                        raise TypeError(
                            "shell_layer attribute must be a core.shell_layers instance."
                        )
                    sl = shell_layers
                changeOp.inputs.e_shell_layer.connect(sl.value)  # top layers taken
                fields_container = changeOp.get_output(0, core.types.fields_container)
                break

        # Merge field data into a single array
        if component_count > 1:
            overall_data = np.full((len(mesh_location), component_count), np.nan)
        else:
            overall_data = np.full(len(mesh_location), np.nan)

        for field in fields_container:
            ind, mask = mesh_location.map_scoping(field.scoping)
            overall_data[ind] = field.data[mask]

        # create the plotter and add the meshes

        # add meshes
        kwargs.setdefault("stitle", name)
        kwargs = self._internal_plotter._set_scalar_bar_title(kwargs)

        kwargs.setdefault("show_edges", True)
        kwargs.setdefault("nan_color", "grey")

        # Set the scalar bar title
        kwargs.setdefault("stitle", f"{name} ({unit})")
        kwargs = self._internal_plotter._set_scalar_bar_title(kwargs)

        # show axes
        show_axes = kwargs.pop("show_axes", None)
        if show_axes:
            self._internal_plotter._plotter.add_axes()

        text = kwargs.pop("text", None)
        if text is not None:
            self._internal_plotter._plotter.add_text(text, position="lower_edge")

        kwargs_in = _sort_supported_kwargs(
            bound_method=self._internal_plotter._plotter.add_mesh, **kwargs
        )
        as_linear = True
        if deform_by:
            grid = mesh._as_vtk(mesh.deform_by(deform_by, scale_factor), as_linear=as_linear)
            self._internal_plotter.add_scale_factor_legend(scale_factor, **kwargs)
        else:
            if as_linear != mesh.as_linear:
                grid = mesh._as_vtk(mesh.nodes.coordinates_field, as_linear=as_linear)
                mesh.as_linear = as_linear
            else:
                grid = mesh.grid
        grid.clear_data()
        self._internal_plotter._plotter.add_mesh(grid, scalars=overall_data, **kwargs_in)

        background = kwargs.pop("background", None)
        if background is not None:
            self._internal_plotter._plotter.set_background(background)

        cpos = kwargs.pop("cpos", None)
        if cpos is not None:
            self._internal_plotter._plotter.camera_position = cpos

        # show result
        kwargs_in = _sort_supported_kwargs(
            bound_method=self._internal_plotter._plotter.show, **kwargs
        )
        return self._internal_plotter._plotter.show(**kwargs_in)

    def _plot_contour_using_vtk_file(self, fields_container, notebook=None):
        """Plot the contour result on its mesh support.

        The resulting figure depends on the support, which can be a meshed region
        or a time freq support. If a transient analysis, the last result is plotted.

        This method is private.  DPF publishes a VTK file and displays
        this file using PyVista.
        """
        try:
            import pyvista as pv
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To use plotting capabilities, please install pyvista "
                "with :\n pip install pyvista>=0.24.0"
            )

        plotter = pv.Plotter(notebook=notebook)
        # mesh_provider = Operator("MeshProvider")
        # mesh_provider.inputs.data_sources.connect(self._evaluator._model.metadata.data_sources)

        # create a temporary file at the default temp directory
        path = os.path.join(tempfile.gettempdir(), "dpf_temp_hokflb2j9s.vtk")

        vtk_export = dpf.core.Operator("vtk_export")
        vtk_export.inputs.mesh.connect(self._mesh)
        vtk_export.inputs.fields1.connect(fields_container)
        vtk_export.inputs.file_path.connect(path)
        vtk_export.run()
        grid = pv.read(path)

        if os.path.exists(path):
            os.remove(path)

        names = grid.array_names
        field_name = fields_container[0].name
        for n in names:  # get new name (for example if time_steps)
            if field_name in n:
                field_name = n  # default: will plot the last time_step
        val = grid.get_array(field_name)
        plotter.add_mesh(grid, scalars=val, stitle=field_name, show_edges=True)
        plotter.add_axes()
        plotter.show()



"""
.. _ref_data_tree:

DataTree
========
"""
import enum
import traceback
import warnings
import weakref

from ansys.dpf.core.mapping_types import types
from ansys.dpf.core import server as server_module
from ansys.dpf.core import collection
from ansys.dpf.core import errors
from ansys.dpf.gate import (
    dpf_data_tree_abstract_api,
    dpf_data_tree_capi,
    dpf_data_tree_grpcapi,
    data_processing_capi,
    data_processing_grpcapi,
    integral_types,
)


class DataTree:
    """Represents an entity mapping attributes names to values.

    Parameters
    ----------
    data: dict(string:object)
        Dictionary attributes names to its associated data to add to the data tree.
    data_tree : ctypes.c_void_p, ansys.grpc.dpf.data_tree_pb2.DataTree message, optional
    server : DPFServer, optional
        Server with channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Examples
    --------
    Create a data tree from a dictionary.

    >>> from ansys.dpf import core as dpf
    >>> data_tree = dpf.DataTree({"num_entities":3, "list_of_raws":[1,2,3,4], "name": "George"})
    >>> data_tree.get_as("name", dpf.types.string)
    'George'
    >>> data_tree.get_as("list_of_raws", dpf.types.vec_int)
    [1, 2, 3, 4]

    Create a data tree with add.

    >>> from ansys.dpf import core as dpf
    >>> data_tree = dpf.DataTree()
    >>> data_tree.add(num_entities=3, list_of_raws=[1,2,3,4], name="George")
    >>> txt = data_tree.write_to_txt()

    Create a data tree with a context manager.

    >>> from ansys.dpf import core as dpf
    >>> data_tree = dpf.DataTree()
    >>> with data_tree.to_fill() as to_fill:
    ...     to_fill.num_entities = 3
    ...     to_fill.list_of_raws = [1,2,3,4]
    >>> json = data_tree.write_to_json()

    Notes
    -----
    Class available with server's version starting at 4.0 (Ansys 2022R2).
    """

    def __init__(self, data=None, data_tree=None, server=None):
        # __set_attr__ method has been overridden, self._common_keys is used to list the "real"
        # names used as its class attributes
        self._common_keys = [
            "_common_keys",
            "_server",
            "_internal_obj",
            "_owner_data_tree",
            "_dict",
            "_api_instance",
            "_deleter_func",
            "_holds_server_instance",
            "_server_instance",
            "_holds_server",
        ]
        # step 1: get server
        self._server_instance = server_module.get_or_create_server(server)

        if data_tree is None and not self._server.meet_version("4.0"):
            raise errors.DpfVersionNotSupported("4.0")

        # step 2: get api
        self._api_instance = None  # see property self._api

        # step4: if object exists, take the instance, else create it
        if data_tree is not None:
            if not isinstance(data_tree, dict):
                # step3: init environment
                self._api.init_dpf_data_tree_environment(self)  # creates stub when gRPC
            self._internal_obj = data_tree
        else:
            # step3: init environment
            self._api.init_dpf_data_tree_environment(self)  # creates stub when gRPC
            if self._server.has_client():
                self._internal_obj = self._api.dpf_data_tree_new_on_client(self._server.client)
            else:
                self._internal_obj = self._api.dpf_data_tree_new()

        if data:
            self.add(data)

        self._holds_server_instance = True

    @property
    def _holds_server(self):
        return self._holds_server_instance

    @property
    def _server(self):
        if isinstance(self._server_instance, weakref.ref):
            return self._server_instance()
        return self._server_instance

    @_holds_server.setter
    def _holds_server(self, value):
        self._holds_server_instance = value
        if value is False and not isinstance(self._server_instance, weakref.ref):
            self._server_instance = weakref.ref(self._server_instance)

    @property
    def _api(self) -> dpf_data_tree_abstract_api.DpfDataTreeAbstractAPI:
        if not self._api_instance:
            self._api_instance = self._server.get_api_for_type(
                capi=dpf_data_tree_capi.DpfDataTreeCAPI,
                grpcapi=dpf_data_tree_grpcapi.DpfDataTreeGRPCAPI,
            )
        return self._api_instance

    def add(self, *args, **kwargs):
        """
        Add attributes with their value to the data tree.

        Parameters
        ----------
        args : dict[string:object], optional
        kwargs : int, float, string, list[int], list[double], list[str], DataTree
            Attributes names and their values to add.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> data_tree.add(id=3, qualities=["nice", "funny"], name="George")

        """

        def add_data(self, key, value):
            if isinstance(value, str):
                self._api.dpf_data_tree_set_string_attribute(self, key, value, len(value))
            elif isinstance(value, float):
                self._api.dpf_data_tree_set_double_attribute(self, key, value)
            elif isinstance(value, (bool, int, enum.Enum)):
                self._api.dpf_data_tree_set_int_attribute(self, key, int(value))
            elif isinstance(value, list):
                if len(value) > 0 and isinstance(value[0], float):
                    self._api.dpf_data_tree_set_vec_double_attribute(self, key, value, len(value))
                elif len(value) > 0 and isinstance(value[0], str):
                    coll_obj = collection.StringCollection(
                        list=value, local=True, server=self._server
                    )
                    self._api.dpf_data_tree_set_string_collection_attribute(self, key, coll_obj)
                elif len(value) > 0 and isinstance(value[0], int):
                    self._api.dpf_data_tree_set_vec_int_attribute(self, key, value, len(value))
                elif len(value) > 0:
                    raise TypeError(
                        f"List of {type(value[0]).__name__} is not supported, "
                        f"use list of int, float or strings."
                    )
            elif isinstance(value, DataTree):
                self._api.dpf_data_tree_set_sub_tree_attribute(self, key, value)
            else:
                raise TypeError(
                    f"{type(value[0]).__name__} is not a supported type, "
                    "use lists, int, float, strings or DataTree."
                )

        for entry in args:
            for key, value in entry.items():
                add_data(self, key, value)

        for key, value in kwargs.items():
            add_data(self, key, value)

    @property
    def _core_api(self):
        core_api = self._server.get_api_for_type(
            capi=data_processing_capi.DataProcessingCAPI,
            grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
        )
        core_api.init_data_processing_environment(self)
        return core_api

    def to_fill(self):
        """
        This method allows to access and modify the local copy of the data_tree
        without sending a request to the server. It should be used in a ``with``
        statement so that the local data tree is released and the data is sent to
        the server in one action. If it is not used in a ``with`` statement,
        :func:`<release_data> DataTree.release_data()` should be used to update the data tree.

        Warning
        -------
        If this :func:`<release_data> DataTree.to_fill()` method is not used as
        a context manager in a ``with`` statement or if the method `release_data()`
        is not called,
        the data will not be updated.

        Returns
        -------
        local_data_tree : DataTree

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> with data_tree.to_fill() as to_fill:
        ...     to_fill.name = "George"
        ...     to_fill.qualities=["nice", "funny"]
        ...     to_fill.id = 3

        """
        return _LocalDataTree(self)

    def _serialize(self, path, operator):
        from ansys.dpf import core

        operator.inputs.data_tree.connect(self)
        if path:
            if self._server.local_server:
                operator.inputs.path.connect(path)
                operator.run()
                return path
            else:
                directory = core.core.make_tmp_dir_server(self._server)
                server_path = core.path_utilities.join(directory, "tmp.txt", server=self._server)
                operator.inputs.path.connect(server_path)
                operator.run()
                return core.download_file(server_path, path, server=self._server)
        else:
            return operator.get_output(0, core.types.string)

    def write_to_txt(self, path=None):
        """
        Writes the data tree either as a file or as returned string in a text format.

        Parameters
        ----------
        path : str, optional
            If a path is specified the output is written to this file.

        Returns
        -------
        str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> data_tree.add(id=3, qualities=["nice", "funny"], name="George")
        >>> txt = data_tree.write_to_txt()
        >>> import tempfile
        >>> import os
        >>> data_tree.write_to_txt(os.path.join(tempfile.mkdtemp(), "data_tree.txt"))
        <BLANKLINE>
        ...

        """
        from ansys.dpf.core.operators.serialization import data_tree_to_txt

        op = data_tree_to_txt(server=self._server)
        return self._serialize(path, op)

    def write_to_json(self, path=None):
        """
        Writes the data tree either as a file or as returned string in a json format.

        Parameters
        ----------
        path : str, optional
            If a path is specified the output is written to this file.

        Returns
        -------
        str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> data_tree.add(id=3, qualities=["nice", "funny"], name="George")
        >>> txt = data_tree.write_to_json()
        >>> import tempfile
        >>> import os
        >>> data_tree.write_to_json(os.path.join(tempfile.mkdtemp(), "data_tree.json"))
        <BLANKLINE>
        ...

        """
        from ansys.dpf.core.operators.serialization import data_tree_to_json

        op = data_tree_to_json(server=self._server)
        return self._serialize(path, op)

    @staticmethod
    def _deserialize(path, txt, server, operator):
        from ansys.dpf import core

        if path:
            server = server_module.get_or_create_server(server)
            if server.local_server:
                operator.inputs.string_or_path.connect(core.DataSources(path, server=server))
            else:
                server_path = core.upload_file_in_tmp_folder(path, server=server)
                operator.inputs.string_or_path.connect(core.DataSources(server_path, server=server))
        elif txt:
            operator.inputs.string_or_path.connect(str(txt))
        return operator.outputs.data_tree()

    @staticmethod
    def read_from_json(path=None, txt=None, server=None):
        """
        Convert a json string or file to DataTree

        Parameters
        ----------
        path : str, optional
            If a path is specified the output is read from this file.

        txt : str, optional
            If a txt is specified the output is read from this string.

        server : DPFServer, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        DataTree

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree({"num_entities":3, "list_of_raws":[1,2,3,4], "name": "George"})
        >>> txt = data_tree.write_to_json()
        >>> data_tree_copy = dpf.DataTree.read_from_json(txt=txt)

        """
        from ansys.dpf.core.operators.serialization import json_to_data_tree

        op = json_to_data_tree(server=server)
        return DataTree._deserialize(path, txt, server, op)

    @staticmethod
    def read_from_txt(path=None, txt=None, server=None):
        """
        Convert a text string or file to DataTree

        Parameters
        ----------
        path : str, optional
            If a path is specified the output is read from this file.

        txt : str, optional
            If a txt is specified the output is read from this string.

        server : DPFServer, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        DataTree

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree({"num_entities":3, "list_of_raws":[1,2,3,4], "name": "George"})
        >>> txt = data_tree.write_to_txt()
        >>> data_tree_copy = dpf.DataTree.read_from_txt(txt=txt)

        """
        from ansys.dpf.core.operators.serialization import txt_to_data_tree

        op = txt_to_data_tree(server=server)
        return DataTree._deserialize(path, txt, server, op)

    def has(self, entry):
        """
        Return True if the entry exists

        Parameters
        ----------
        entry : str

        Returns
        -------
        bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> data_tree.add(id=3, qualities=["nice", "funny"], name="George")
        >>> data_tree.has("qualities")
        True
        >>> data_tree.has("flaws")
        False

        """
        return self._api.dpf_data_tree_has_attribute(self, entry)

    def get_as(self, name, type_to_return=types.string):
        """
        Returns an attribute value by its name in the required type.

        Parameters
        ----------
        name : str
            Name of the attribute to return
        type_to_return : types
            Type of the attribute to return. String is supported for all attributes.

        Returns
        -------
        str, int, float, list[int], list[float], list[str], DataTree

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> data_tree.add(id=3, qualities=["nice", "funny"], name="George")
        >>> data_tree.get_as("id")
        '3'
        >>> data_tree.get_as("id", dpf.types.int)
        3
        >>> data_tree.get_as("qualities", dpf.types.vec_string)
        ['nice', 'funny']

        """
        out = None
        if type_to_return == types.int:
            out = integral_types.MutableInt32()
            self._api.dpf_data_tree_get_int_attribute(self, name, out)
            out = int(out)
        elif type_to_return == types.double:
            out = integral_types.MutableDouble()
            self._api.dpf_data_tree_get_double_attribute(self, name, out)
            out = float(out)
        elif type_to_return == types.string:
            out = integral_types.MutableString(1)
            size = integral_types.MutableInt32(0)
            self._api.dpf_data_tree_get_string_attribute(self, name, out, size)
            out = str(out)
        elif type_to_return == types.vec_double:
            out = integral_types.MutableListDouble()
            self._api.dpf_data_tree_get_vec_double_attribute(self, name, out, out.internal_size)
            out = out.tolist()
        elif type_to_return == types.vec_int:
            out = integral_types.MutableListInt32()
            self._api.dpf_data_tree_get_vec_int_attribute(self, name, out, out.internal_size)
            out = out.tolist()
        elif type_to_return == types.vec_string:
            coll_obj = collection.StringCollection(
                collection=self._api.dpf_data_tree_get_string_collection_attribute(self, name),
                server=self._server,
            )
            out = coll_obj.get_integral_entries()
        elif type_to_return == types.data_tree:
            obj = self._api.dpf_data_tree_get_sub_tree(self, name)
            out = DataTree(data_tree=obj, server=self._server)
        return out

    @property
    def attribute_names(self):
        """
        Returns a list of defined attribute names.

        Returns
        -------
        list[str]

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> data_tree.add(id=3, qualities=["nice", "funny"], name="George")
        >>> data_tree.attribute_names
        ['id', 'name', 'qualities']
        """
        coll_obj = collection.StringCollection(
            collection=self._api.dpf_data_tree_get_available_attributes_names_in_string_collection(
                self
            ),
            server=self._server,
        )

        return coll_obj.get_integral_entries()

    @property
    def sub_tree_names(self):
        """
        Returns a list of defined sub-tree names.

        Returns
        -------
        list[str]

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> first_subtree = dpf.DataTree()
        >>> second_subtree = dpf.DataTree()
        >>> data_tree.add(first=first_subtree, second=second_subtree)
        >>> data_tree.sub_tree_names
        ['first', 'second']
        """
        coll_obj = collection.StringCollection(
            collection=self._api.dpf_data_tree_get_available_sub_tree_names_in_string_collection(
                self
            ),
            server=self._server,
        )

        return coll_obj.get_integral_entries()

    @attribute_names.setter
    def attribute_names(self, val):
        raise AttributeError("can't set attribute")

    @sub_tree_names.setter
    def sub_tree_names(self, val):
        raise AttributeError("can't set attribute")

    def __to_dict(self, dic):
        for attribute_name in self.attribute_names:
            dic[attribute_name] = self.get_as(attribute_name)

        for sub_tree_name in self.sub_tree_names:
            sub_tree = self.get_as(sub_tree_name, types.data_tree)
            sub_dic = {}
            sub_tree.__to_dict(sub_dic)
            dic[sub_tree_name] = sub_dic

    def to_dict(self):
        """
        Returns a read-only dictionary representation of the DataTree.

        Returns
        -------
        dict

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_tree = dpf.DataTree()
        >>> sub = dpf.DataTree()
        >>> sub.add(str="hello world")
        >>> data_tree.add(id=3, sub_tree=sub)
        >>> data_tree.to_dict()
        {'id': '3', 'sub_tree': {'str': 'hello world'}}
        """
        dic = {}
        self.__to_dict(dic)

        return dic

    def __setattr__(self, key, value):
        if key == "_common_keys" or key in self._common_keys or key in dir(self):
            return super.__setattr__(self, key, value)
        self.add({key: value})

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        str
            Description of the entity.
        """
        from ansys.dpf.core.core import _description

        return _description(self._internal_obj, self._server)

    def __del__(self):
        try:
            # needs a proper deleter only when real datatree and not dict
            if hasattr(self, "_deleter_func"):
                obj = self._deleter_func[1](self)
                if obj is not None:
                    self._deleter_func[0](obj)
        except:
            warnings.warn(traceback.format_exc())


class _LocalDataTree(DataTree):
    def __init__(self, data_tree):
        self._common_keys = ["_owner_data_tree", "_dict", "_common_keys", "_is_exited"]
        self._owner_data_tree = data_tree
        self.__cache_data__()
        self._is_exited = False

    def add(self, *args, **kwargs):
        """
        Add attributes with their value to the data tree.

        Parameters
        ----------
        kwargs: int, float, string, list[int], list[double], list[str], DataTree
            Data to add.
        """

        def add_data(key, value):
            self._dict[key] = value

        for entry in args:
            for key, value in entry.items():
                add_data(key, value)
        for key, value in kwargs.items():
            add_data(key, value)

    def __cache_data__(self):
        self._dict = {}

    def release_data(self):
        """Release the data."""
        self._owner_data_tree.add(self._dict)

    def __setattr__(self, key, value):
        if key == "_common_keys" or key in self._common_keys:
            return super.__setattr__(self, key, value)
        self.add({key: value})

    def __enter__(self):
        return self

    def __exit__(self, type_to_use, value, tb):
        if tb is None:
            self._is_exited = True
            self.release_data()
        else:
            print(tb)

    def __del__(self):
        if not self._is_exited:
            self._is_exited = True
            self.release_data()



"""
extract_time_freq
=================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.outputs import _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class extract_time_freq(Operator):
    """Extract timefreqs with sets scoping from a time freq support

    Parameters
    ----------
    time_freq_support : TimeFreqSupport
    time_scoping : Scoping, optional
    real_or_complex : bool, optional
        False for real only (default). true for
        complex output.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.extract_time_freq()

    >>> # Make input connections
    >>> my_time_freq_support = dpf.TimeFreqSupport()
    >>> op.inputs.time_freq_support.connect(my_time_freq_support)
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_real_or_complex = bool()
    >>> op.inputs.real_or_complex.connect(my_real_or_complex)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.extract_time_freq(
    ...     time_freq_support=my_time_freq_support,
    ...     time_scoping=my_time_scoping,
    ...     real_or_complex=my_real_or_complex,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(
        self,
        time_freq_support=None,
        time_scoping=None,
        real_or_complex=None,
        config=None,
        server=None,
    ):
        super().__init__(name="extract_time_freq", config=config, server=server)
        self._inputs = InputsExtractTimeFreq(self)
        self._outputs = OutputsExtractTimeFreq(self)
        if time_freq_support is not None:
            self.inputs.time_freq_support.connect(time_freq_support)
        if time_scoping is not None:
            self.inputs.time_scoping.connect(time_scoping)
        if real_or_complex is not None:
            self.inputs.real_or_complex.connect(real_or_complex)

    @staticmethod
    def _spec():
        description = """Extract timefreqs with sets scoping from a time freq support"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_freq_support",
                    type_names=["time_freq_support"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="time_scoping",
                    type_names=["scoping"],
                    optional=True,
                    document="""""",
                ),
                2: PinSpecification(
                    name="real_or_complex",
                    type_names=["bool"],
                    optional=True,
                    document="""False for real only (default). true for
        complex output.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="extract_time_freq", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsExtractTimeFreq
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsExtractTimeFreq
        """
        return super().outputs


class InputsExtractTimeFreq(_Inputs):
    """Intermediate class used to connect user inputs to
    extract_time_freq operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.extract_time_freq()
    >>> my_time_freq_support = dpf.TimeFreqSupport()
    >>> op.inputs.time_freq_support.connect(my_time_freq_support)
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_real_or_complex = bool()
    >>> op.inputs.real_or_complex.connect(my_real_or_complex)
    """

    def __init__(self, op: Operator):
        super().__init__(extract_time_freq._spec().inputs, op)
        self._time_freq_support = Input(
            extract_time_freq._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._time_freq_support)
        self._time_scoping = Input(extract_time_freq._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._time_scoping)
        self._real_or_complex = Input(extract_time_freq._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._real_or_complex)

    @property
    def time_freq_support(self):
        """Allows to connect time_freq_support input to the operator.

        Parameters
        ----------
        my_time_freq_support : TimeFreqSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.extract_time_freq()
        >>> op.inputs.time_freq_support.connect(my_time_freq_support)
        >>> # or
        >>> op.inputs.time_freq_support(my_time_freq_support)
        """
        return self._time_freq_support

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator.

        Parameters
        ----------
        my_time_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.extract_time_freq()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> # or
        >>> op.inputs.time_scoping(my_time_scoping)
        """
        return self._time_scoping

    @property
    def real_or_complex(self):
        """Allows to connect real_or_complex input to the operator.

        False for real only (default). true for
        complex output.

        Parameters
        ----------
        my_real_or_complex : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.extract_time_freq()
        >>> op.inputs.real_or_complex.connect(my_real_or_complex)
        >>> # or
        >>> op.inputs.real_or_complex(my_real_or_complex)
        """
        return self._real_or_complex


class OutputsExtractTimeFreq(_Outputs):
    """Intermediate class used to get outputs from
    extract_time_freq operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.extract_time_freq()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(extract_time_freq._spec().outputs, op)
        self.field_as_field = Output(
            _modify_output_spec_with_one_type(
                extract_time_freq._spec().output_pin(0), "field"
            ),
            0,
            op,
        )
        self._outputs.append(self.field_as_field)
        self.field_as_fields_container = Output(
            _modify_output_spec_with_one_type(
                extract_time_freq._spec().output_pin(0), "fields_container"
            ),
            0,
            op,
        )
        self._outputs.append(self.field_as_fields_container)



"""
extract_field
=============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class extract_field(Operator):
    """Extracts the fields at the indices defined in the vector (in 1) from
    the fields container (in 0).

    Parameters
    ----------
    fields_container : Field or FieldsContainer
        If a field is in input, it is passed on as an
        output
    indices : optional
        Default is the first field


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.extract_field()

    >>> # Make input connections
    >>> my_fields_container = dpf.Field()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_indices = dpf.()
    >>> op.inputs.indices.connect(my_indices)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.extract_field(
    ...     fields_container=my_fields_container,
    ...     indices=my_indices,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(self, fields_container=None, indices=None, config=None, server=None):
        super().__init__(name="ExtractFromFC", config=config, server=server)
        self._inputs = InputsExtractField(self)
        self._outputs = OutputsExtractField(self)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if indices is not None:
            self.inputs.indices.connect(indices)

    @staticmethod
    def _spec():
        description = """Extracts the fields at the indices defined in the vector (in 1) from
            the fields container (in 0)."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""If a field is in input, it is passed on as an
        output""",
                ),
                1: PinSpecification(
                    name="indices",
                    type_names=["vector<int32>"],
                    optional=True,
                    document="""Default is the first field""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="ExtractFromFC", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsExtractField
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsExtractField
        """
        return super().outputs


class InputsExtractField(_Inputs):
    """Intermediate class used to connect user inputs to
    extract_field operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.extract_field()
    >>> my_fields_container = dpf.Field()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_indices = dpf.()
    >>> op.inputs.indices.connect(my_indices)
    """

    def __init__(self, op: Operator):
        super().__init__(extract_field._spec().inputs, op)
        self._fields_container = Input(extract_field._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._fields_container)
        self._indices = Input(extract_field._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._indices)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        If a field is in input, it is passed on as an
        output

        Parameters
        ----------
        my_fields_container : Field or FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.extract_field()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def indices(self):
        """Allows to connect indices input to the operator.

        Default is the first field

        Parameters
        ----------
        my_indices :

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.extract_field()
        >>> op.inputs.indices.connect(my_indices)
        >>> # or
        >>> op.inputs.indices(my_indices)
        """
        return self._indices


class OutputsExtractField(_Outputs):
    """Intermediate class used to get outputs from
    extract_field operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.extract_field()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(extract_field._spec().outputs, op)
        self._field = Output(extract_field._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.extract_field()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



"""
unit_convert
============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.outputs import _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class unit_convert(Operator):
    """Converts an input field/fields container or mesh of a given unit to
    another unit.

    Parameters
    ----------
    entity_to_convert : Field or FieldsContainer or MeshedRegion or MeshesContainer
    unit_name : str or int
        Unit as a string, ex 'm' for meter, 'pa' for
        pascal,... or ansys unit system's id


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.math.unit_convert()

    >>> # Make input connections
    >>> my_entity_to_convert = dpf.Field()
    >>> op.inputs.entity_to_convert.connect(my_entity_to_convert)
    >>> my_unit_name = str()
    >>> op.inputs.unit_name.connect(my_unit_name)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.math.unit_convert(
    ...     entity_to_convert=my_entity_to_convert,
    ...     unit_name=my_unit_name,
    ... )

    >>> # Get output data
    >>> result_converted_entity = op.outputs.converted_entity()
    """

    def __init__(
        self, entity_to_convert=None, unit_name=None, config=None, server=None
    ):
        super().__init__(name="unit_convert", config=config, server=server)
        self._inputs = InputsUnitConvert(self)
        self._outputs = OutputsUnitConvert(self)
        if entity_to_convert is not None:
            self.inputs.entity_to_convert.connect(entity_to_convert)
        if unit_name is not None:
            self.inputs.unit_name.connect(unit_name)

    @staticmethod
    def _spec():
        description = """Converts an input field/fields container or mesh of a given unit to
            another unit."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="entity_to_convert",
                    type_names=[
                        "field",
                        "fields_container",
                        "abstract_meshed_region",
                        "meshes_container",
                    ],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="unit_name",
                    type_names=["string", "int32"],
                    optional=False,
                    document="""Unit as a string, ex 'm' for meter, 'pa' for
        pascal,... or ansys unit system's id""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="converted_entity",
                    type_names=[
                        "field",
                        "fields_container",
                        "abstract_meshed_region",
                        "meshes_container",
                    ],
                    optional=False,
                    document="""The output entity is the same as the input
        (inplace operator)""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="unit_convert", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsUnitConvert
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsUnitConvert
        """
        return super().outputs


class InputsUnitConvert(_Inputs):
    """Intermediate class used to connect user inputs to
    unit_convert operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.unit_convert()
    >>> my_entity_to_convert = dpf.Field()
    >>> op.inputs.entity_to_convert.connect(my_entity_to_convert)
    >>> my_unit_name = str()
    >>> op.inputs.unit_name.connect(my_unit_name)
    """

    def __init__(self, op: Operator):
        super().__init__(unit_convert._spec().inputs, op)
        self._entity_to_convert = Input(unit_convert._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._entity_to_convert)
        self._unit_name = Input(unit_convert._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._unit_name)

    @property
    def entity_to_convert(self):
        """Allows to connect entity_to_convert input to the operator.

        Parameters
        ----------
        my_entity_to_convert : Field or FieldsContainer or MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.unit_convert()
        >>> op.inputs.entity_to_convert.connect(my_entity_to_convert)
        >>> # or
        >>> op.inputs.entity_to_convert(my_entity_to_convert)
        """
        return self._entity_to_convert

    @property
    def unit_name(self):
        """Allows to connect unit_name input to the operator.

        Unit as a string, ex 'm' for meter, 'pa' for
        pascal,... or ansys unit system's id

        Parameters
        ----------
        my_unit_name : str or int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.unit_convert()
        >>> op.inputs.unit_name.connect(my_unit_name)
        >>> # or
        >>> op.inputs.unit_name(my_unit_name)
        """
        return self._unit_name


class OutputsUnitConvert(_Outputs):
    """Intermediate class used to get outputs from
    unit_convert operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.unit_convert()
    >>> # Connect inputs : op.inputs. ...
    >>> result_converted_entity = op.outputs.converted_entity()
    """

    def __init__(self, op: Operator):
        super().__init__(unit_convert._spec().outputs, op)
        self.converted_entity_as_field = Output(
            _modify_output_spec_with_one_type(
                unit_convert._spec().output_pin(0), "field"
            ),
            0,
            op,
        )
        self._outputs.append(self.converted_entity_as_field)
        self.converted_entity_as_fields_container = Output(
            _modify_output_spec_with_one_type(
                unit_convert._spec().output_pin(0), "fields_container"
            ),
            0,
            op,
        )
        self._outputs.append(self.converted_entity_as_fields_container)
        self.converted_entity_as_meshed_region = Output(
            _modify_output_spec_with_one_type(
                unit_convert._spec().output_pin(0), "meshed_region"
            ),
            0,
            op,
        )
        self._outputs.append(self.converted_entity_as_meshed_region)
        self.converted_entity_as_meshes_container = Output(
            _modify_output_spec_with_one_type(
                unit_convert._spec().output_pin(0), "meshes_container"
            ),
            0,
            op,
        )
        self._outputs.append(self.converted_entity_as_meshes_container)



"""
python_generator
================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class python_generator(Operator):
    """Generates .py file with specifications for loaded plugin(s).

    Parameters
    ----------
    dll_source_path : str
    output_path : str
    load_symbol : str, optional
    library_key : str, optional


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.python_generator()

    >>> # Make input connections
    >>> my_dll_source_path = str()
    >>> op.inputs.dll_source_path.connect(my_dll_source_path)
    >>> my_output_path = str()
    >>> op.inputs.output_path.connect(my_output_path)
    >>> my_load_symbol = str()
    >>> op.inputs.load_symbol.connect(my_load_symbol)
    >>> my_library_key = str()
    >>> op.inputs.library_key.connect(my_library_key)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.python_generator(
    ...     dll_source_path=my_dll_source_path,
    ...     output_path=my_output_path,
    ...     load_symbol=my_load_symbol,
    ...     library_key=my_library_key,
    ... )

    """

    def __init__(
        self,
        dll_source_path=None,
        output_path=None,
        load_symbol=None,
        library_key=None,
        config=None,
        server=None,
    ):
        super().__init__(name="python_generator", config=config, server=server)
        self._inputs = InputsPythonGenerator(self)
        self._outputs = OutputsPythonGenerator(self)
        if dll_source_path is not None:
            self.inputs.dll_source_path.connect(dll_source_path)
        if output_path is not None:
            self.inputs.output_path.connect(output_path)
        if load_symbol is not None:
            self.inputs.load_symbol.connect(load_symbol)
        if library_key is not None:
            self.inputs.library_key.connect(library_key)

    @staticmethod
    def _spec():
        description = """Generates .py file with specifications for loaded plugin(s)."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="dll_source_path",
                    type_names=["string"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="output_path",
                    type_names=["string"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="load_symbol",
                    type_names=["string"],
                    optional=True,
                    document="""""",
                ),
                3: PinSpecification(
                    name="library_key",
                    type_names=["string"],
                    optional=True,
                    document="""""",
                ),
            },
            map_output_pin_spec={},
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="python_generator", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPythonGenerator
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsPythonGenerator
        """
        return super().outputs


class InputsPythonGenerator(_Inputs):
    """Intermediate class used to connect user inputs to
    python_generator operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.python_generator()
    >>> my_dll_source_path = str()
    >>> op.inputs.dll_source_path.connect(my_dll_source_path)
    >>> my_output_path = str()
    >>> op.inputs.output_path.connect(my_output_path)
    >>> my_load_symbol = str()
    >>> op.inputs.load_symbol.connect(my_load_symbol)
    >>> my_library_key = str()
    >>> op.inputs.library_key.connect(my_library_key)
    """

    def __init__(self, op: Operator):
        super().__init__(python_generator._spec().inputs, op)
        self._dll_source_path = Input(python_generator._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._dll_source_path)
        self._output_path = Input(python_generator._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._output_path)
        self._load_symbol = Input(python_generator._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._load_symbol)
        self._library_key = Input(python_generator._spec().input_pin(3), 3, op, -1)
        self._inputs.append(self._library_key)

    @property
    def dll_source_path(self):
        """Allows to connect dll_source_path input to the operator.

        Parameters
        ----------
        my_dll_source_path : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.python_generator()
        >>> op.inputs.dll_source_path.connect(my_dll_source_path)
        >>> # or
        >>> op.inputs.dll_source_path(my_dll_source_path)
        """
        return self._dll_source_path

    @property
    def output_path(self):
        """Allows to connect output_path input to the operator.

        Parameters
        ----------
        my_output_path : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.python_generator()
        >>> op.inputs.output_path.connect(my_output_path)
        >>> # or
        >>> op.inputs.output_path(my_output_path)
        """
        return self._output_path

    @property
    def load_symbol(self):
        """Allows to connect load_symbol input to the operator.

        Parameters
        ----------
        my_load_symbol : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.python_generator()
        >>> op.inputs.load_symbol.connect(my_load_symbol)
        >>> # or
        >>> op.inputs.load_symbol(my_load_symbol)
        """
        return self._load_symbol

    @property
    def library_key(self):
        """Allows to connect library_key input to the operator.

        Parameters
        ----------
        my_library_key : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.python_generator()
        >>> op.inputs.library_key.connect(my_library_key)
        >>> # or
        >>> op.inputs.library_key(my_library_key)
        """
        return self._library_key


class OutputsPythonGenerator(_Outputs):
    """Intermediate class used to get outputs from
    python_generator operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.python_generator()
    >>> # Connect inputs : op.inputs. ...
    """

    def __init__(self, op: Operator):
        super().__init__(python_generator._spec().outputs, op)



"""
make_label_space
================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class make_label_space(Operator):
    """Assemble strings and integers to make a label space.

    Parameters
    ----------
    base_label : dict or FieldsContainer or ScopingsContainer, optional
        Used as a base label (extracted from
        fields/scoping container, or directly
        from label space) that is
        concatenated with provided values.
    label_name : str
    label_value1 : int
    label_value2 : int


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.make_label_space()

    >>> # Make input connections
    >>> my_base_label = dict()
    >>> op.inputs.base_label.connect(my_base_label)
    >>> my_label_name = str()
    >>> op.inputs.label_name.connect(my_label_name)
    >>> my_label_value1 = int()
    >>> op.inputs.label_value1.connect(my_label_value1)
    >>> my_label_value2 = int()
    >>> op.inputs.label_value2.connect(my_label_value2)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.make_label_space(
    ...     base_label=my_base_label,
    ...     label_name=my_label_name,
    ...     label_value1=my_label_value1,
    ...     label_value2=my_label_value2,
    ... )

    >>> # Get output data
    >>> result_label = op.outputs.label()
    """

    def __init__(
        self,
        base_label=None,
        label_name=None,
        label_value1=None,
        label_value2=None,
        config=None,
        server=None,
    ):
        super().__init__(name="make_label_space", config=config, server=server)
        self._inputs = InputsMakeLabelSpace(self)
        self._outputs = OutputsMakeLabelSpace(self)
        if base_label is not None:
            self.inputs.base_label.connect(base_label)
        if label_name is not None:
            self.inputs.label_name.connect(label_name)
        if label_value1 is not None:
            self.inputs.label_value1.connect(label_value1)
        if label_value2 is not None:
            self.inputs.label_value2.connect(label_value2)

    @staticmethod
    def _spec():
        description = """Assemble strings and integers to make a label space."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="base_label",
                    type_names=[
                        "label_space",
                        "fields_container",
                        "scopings_container",
                    ],
                    optional=True,
                    document="""Used as a base label (extracted from
        fields/scoping container, or directly
        from label space) that is
        concatenated with provided values.""",
                ),
                1: PinSpecification(
                    name="label_name",
                    type_names=["string"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="label_value",
                    type_names=["int32"],
                    optional=False,
                    document="""""",
                ),
                3: PinSpecification(
                    name="label_value",
                    type_names=["int32"],
                    optional=False,
                    document="""""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="label",
                    type_names=["label_space"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="make_label_space", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMakeLabelSpace
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMakeLabelSpace
        """
        return super().outputs


class InputsMakeLabelSpace(_Inputs):
    """Intermediate class used to connect user inputs to
    make_label_space operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.make_label_space()
    >>> my_base_label = dict()
    >>> op.inputs.base_label.connect(my_base_label)
    >>> my_label_name = str()
    >>> op.inputs.label_name.connect(my_label_name)
    >>> my_label_value1 = int()
    >>> op.inputs.label_value1.connect(my_label_value1)
    >>> my_label_value2 = int()
    >>> op.inputs.label_value2.connect(my_label_value2)
    """

    def __init__(self, op: Operator):
        super().__init__(make_label_space._spec().inputs, op)
        self._base_label = Input(make_label_space._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._base_label)
        self._label_name = Input(make_label_space._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._label_name)
        self._label_value1 = Input(make_label_space._spec().input_pin(2), 2, op, 0)
        self._inputs.append(self._label_value1)
        self._label_value2 = Input(make_label_space._spec().input_pin(3), 3, op, 1)
        self._inputs.append(self._label_value2)

    @property
    def base_label(self):
        """Allows to connect base_label input to the operator.

        Used as a base label (extracted from
        fields/scoping container, or directly
        from label space) that is
        concatenated with provided values.

        Parameters
        ----------
        my_base_label : dict or FieldsContainer or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.make_label_space()
        >>> op.inputs.base_label.connect(my_base_label)
        >>> # or
        >>> op.inputs.base_label(my_base_label)
        """
        return self._base_label

    @property
    def label_name(self):
        """Allows to connect label_name input to the operator.

        Parameters
        ----------
        my_label_name : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.make_label_space()
        >>> op.inputs.label_name.connect(my_label_name)
        >>> # or
        >>> op.inputs.label_name(my_label_name)
        """
        return self._label_name

    @property
    def label_value1(self):
        """Allows to connect label_value1 input to the operator.

        Parameters
        ----------
        my_label_value1 : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.make_label_space()
        >>> op.inputs.label_value1.connect(my_label_value1)
        >>> # or
        >>> op.inputs.label_value1(my_label_value1)
        """
        return self._label_value1

    @property
    def label_value2(self):
        """Allows to connect label_value2 input to the operator.

        Parameters
        ----------
        my_label_value2 : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.make_label_space()
        >>> op.inputs.label_value2.connect(my_label_value2)
        >>> # or
        >>> op.inputs.label_value2(my_label_value2)
        """
        return self._label_value2


class OutputsMakeLabelSpace(_Outputs):
    """Intermediate class used to get outputs from
    make_label_space operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.make_label_space()
    >>> # Connect inputs : op.inputs. ...
    >>> result_label = op.outputs.label()
    """

    def __init__(self, op: Operator):
        super().__init__(make_label_space._spec().outputs, op)
        self._label = Output(make_label_space._spec().output_pin(0), 0, op)
        self._outputs.append(self._label)

    @property
    def label(self):
        """Allows to get label output of the operator

        Returns
        ----------
        my_label : dict

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.make_label_space()
        >>> # Connect inputs : op.inputs. ...
        >>> result_label = op.outputs.label()
        """  # noqa: E501
        return self._label



"""
add_rigid_body_motion
=====================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class add_rigid_body_motion(Operator):
    """Adds a given rigid translation, center and rotation from a
    displacement field. The rotation is given in terms of rotations
    angles. Note that the displacement field has to be in the global
    coordinate system

    Parameters
    ----------
    displacement_field : Field
    translation_field : Field
    rotation_field : Field
    center_field : Field
    mesh : MeshedRegion, optional
        Default is the mesh in the support


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.add_rigid_body_motion()

    >>> # Make input connections
    >>> my_displacement_field = dpf.Field()
    >>> op.inputs.displacement_field.connect(my_displacement_field)
    >>> my_translation_field = dpf.Field()
    >>> op.inputs.translation_field.connect(my_translation_field)
    >>> my_rotation_field = dpf.Field()
    >>> op.inputs.rotation_field.connect(my_rotation_field)
    >>> my_center_field = dpf.Field()
    >>> op.inputs.center_field.connect(my_center_field)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.add_rigid_body_motion(
    ...     displacement_field=my_displacement_field,
    ...     translation_field=my_translation_field,
    ...     rotation_field=my_rotation_field,
    ...     center_field=my_center_field,
    ...     mesh=my_mesh,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(
        self,
        displacement_field=None,
        translation_field=None,
        rotation_field=None,
        center_field=None,
        mesh=None,
        config=None,
        server=None,
    ):
        super().__init__(name="RigidBodyAddition", config=config, server=server)
        self._inputs = InputsAddRigidBodyMotion(self)
        self._outputs = OutputsAddRigidBodyMotion(self)
        if displacement_field is not None:
            self.inputs.displacement_field.connect(displacement_field)
        if translation_field is not None:
            self.inputs.translation_field.connect(translation_field)
        if rotation_field is not None:
            self.inputs.rotation_field.connect(rotation_field)
        if center_field is not None:
            self.inputs.center_field.connect(center_field)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        description = """Adds a given rigid translation, center and rotation from a
            displacement field. The rotation is given in terms of
            rotations angles. Note that the displacement field has to
            be in the global coordinate system"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="displacement_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="translation_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="rotation_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                3: PinSpecification(
                    name="center_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                7: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=True,
                    document="""Default is the mesh in the support""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="RigidBodyAddition", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsAddRigidBodyMotion
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsAddRigidBodyMotion
        """
        return super().outputs


class InputsAddRigidBodyMotion(_Inputs):
    """Intermediate class used to connect user inputs to
    add_rigid_body_motion operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.add_rigid_body_motion()
    >>> my_displacement_field = dpf.Field()
    >>> op.inputs.displacement_field.connect(my_displacement_field)
    >>> my_translation_field = dpf.Field()
    >>> op.inputs.translation_field.connect(my_translation_field)
    >>> my_rotation_field = dpf.Field()
    >>> op.inputs.rotation_field.connect(my_rotation_field)
    >>> my_center_field = dpf.Field()
    >>> op.inputs.center_field.connect(my_center_field)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    """

    def __init__(self, op: Operator):
        super().__init__(add_rigid_body_motion._spec().inputs, op)
        self._displacement_field = Input(
            add_rigid_body_motion._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._displacement_field)
        self._translation_field = Input(
            add_rigid_body_motion._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._translation_field)
        self._rotation_field = Input(
            add_rigid_body_motion._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._rotation_field)
        self._center_field = Input(
            add_rigid_body_motion._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._center_field)
        self._mesh = Input(add_rigid_body_motion._spec().input_pin(7), 7, op, -1)
        self._inputs.append(self._mesh)

    @property
    def displacement_field(self):
        """Allows to connect displacement_field input to the operator.

        Parameters
        ----------
        my_displacement_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.add_rigid_body_motion()
        >>> op.inputs.displacement_field.connect(my_displacement_field)
        >>> # or
        >>> op.inputs.displacement_field(my_displacement_field)
        """
        return self._displacement_field

    @property
    def translation_field(self):
        """Allows to connect translation_field input to the operator.

        Parameters
        ----------
        my_translation_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.add_rigid_body_motion()
        >>> op.inputs.translation_field.connect(my_translation_field)
        >>> # or
        >>> op.inputs.translation_field(my_translation_field)
        """
        return self._translation_field

    @property
    def rotation_field(self):
        """Allows to connect rotation_field input to the operator.

        Parameters
        ----------
        my_rotation_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.add_rigid_body_motion()
        >>> op.inputs.rotation_field.connect(my_rotation_field)
        >>> # or
        >>> op.inputs.rotation_field(my_rotation_field)
        """
        return self._rotation_field

    @property
    def center_field(self):
        """Allows to connect center_field input to the operator.

        Parameters
        ----------
        my_center_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.add_rigid_body_motion()
        >>> op.inputs.center_field.connect(my_center_field)
        >>> # or
        >>> op.inputs.center_field(my_center_field)
        """
        return self._center_field

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        Default is the mesh in the support

        Parameters
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.add_rigid_body_motion()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh


class OutputsAddRigidBodyMotion(_Outputs):
    """Intermediate class used to get outputs from
    add_rigid_body_motion operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.add_rigid_body_motion()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(add_rigid_body_motion._spec().outputs, op)
        self._field = Output(add_rigid_body_motion._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.add_rigid_body_motion()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



"""
.. _ref_scoping:

Scoping
=======
"""

import traceback
import warnings
import ctypes

import numpy as np

from ansys.dpf.core.check_version import version_requires
from ansys.dpf.core.common import locations
from ansys.dpf.core import server as server_module
from ansys.dpf.core import server_types
from ansys.dpf.core.cache import _setter
from ansys.dpf.gate import (
    scoping_capi,
    scoping_grpcapi,
    data_processing_capi,
    data_processing_grpcapi,
    dpf_vector_capi,
    dpf_vector_abstract_api,
    dpf_vector,
    dpf_array,
    utils,
)


class Scoping:
    """Represents a scoping, which is a subset of a model support.

    Parameters
    ----------
    scoping : ctypes.c_void_p, ansys.grpc.dpf.scoping_pb2.Scoping message, optional

    server : DPFServer, optional
        Server with channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
    Examples
    --------
    Create a mesh scoping.

    >>> from ansys.dpf import core as dpf
    >>> # 1. using the mesh_scoping_factory
    >>> from ansys.dpf.core import mesh_scoping_factory
    >>> # a. scoping with elemental location that targets the elements with id 2, 7 and 11
    >>> my_elemental_scoping = mesh_scoping_factory.elemental_scoping([2, 7, 11])
    >>> # b. scoping with nodal location that targets the elements with id 4 and 6
    >>> my_nodal_scoping = mesh_scoping_factory.nodal_scoping([4, 6])
    >>> #2. using the classic API
    >>> my_scoping = dpf.Scoping()
    >>> my_scoping.location = dpf.locations.nodal #optional
    >>> my_scoping.ids = list(range(1,11))

    """

    def __init__(self, scoping=None, server=None, ids=None, location=None):
        """Initializes the scoping with an optional scoping message or
        by connecting to a stub.
        """
        # step 1: get server
        self._server = server_module.get_or_create_server(server)
        self._api = self._server.get_api_for_type(
            capi=scoping_capi.ScopingCAPI, grpcapi=scoping_grpcapi.ScopingGRPCAPI
        )
        # step3: init environment
        self._api.init_scoping_environment(self)  # creates stub when gRPC

        # step2: if object exists, take the instance, else create it
        if scoping is not None:
            if isinstance(scoping, Scoping):
                self._server = scoping._server
                self._api = self._server.get_api_for_type(
                    capi=scoping_capi.ScopingCAPI,
                    grpcapi=scoping_grpcapi.ScopingGRPCAPI,
                )
                # step3: init environment
                self._api.init_scoping_environment(self)  # creates stub when gRPC
                core_api = self._server.get_api_for_type(
                    capi=data_processing_capi.DataProcessingCAPI,
                    grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
                )
                core_api.init_data_processing_environment(self)
                self._internal_obj = core_api.data_processing_duplicate_object_reference(scoping)
            else:
                # scoping is of type protobuf.message or DPFObject*
                self._internal_obj = scoping
        else:
            if self._server.has_client():
                self._internal_obj = self._api.scoping_new_on_client(self._server.client)
            else:
                self._internal_obj = self._api.scoping_new()

        # step5: handle specific calls to set attributes
        if ids is not None:
            self.ids = ids
        if location is not None:
            self.location = location

    def _count(self):
        """
        Returns
        -------
        count : int
            Number of scoping IDs.
        """
        return self._api.scoping_get_size(self)

    def _get_location(self):
        """Retrieve the location of the IDs.

        Returns
        -------
        location : str
            Location of the IDs.
        """
        return self._api.scoping_get_location(self)

    def _set_location(self, loc=locations.nodal):
        """
        Parameters
        ----------
        loc : str or core.locations enum
            Location needed.

        """
        self._api.scoping_set_location(self, loc)

    @version_requires("2.1")
    def _set_ids(self, ids):
        """
        Parameters
        ----------
        ids : list of int
            IDs to set.

        Notes
        -----
        Print a progress bar.
        """
        if isinstance(self._server, server_types.InProcessServer):
            self._api.scoping_resize(self, len(ids))
            ids_ptr = self._api.scoping_get_ids(self, len(ids))
            ctypes.memmove(
                ids_ptr,
                utils.to_int32_ptr(ids),
                len(ids) * ctypes.sizeof(ctypes.c_int32()),
            )
        else:
            self._api.scoping_set_ids(self, ids, len(ids))

    def _get_ids(self, np_array=None):
        """
        Returns
        -------
        ids : list[int], numpy.array (if np_array==True)
            Array of IDs.

        np_array: bool, optional

        Notes
        -----
        Print a progress bar.
        """
        if np_array == None:
            from ansys.dpf.core import settings

            np_array = settings.get_runtime_client_config(self._server).return_arrays
        try:
            vec = dpf_vector.DPFVectorInt(
                client=self._server.client,
                api=self._server.get_api_for_type(
                    capi=dpf_vector_capi.DpfVectorCAPI,
                    grpcapi=dpf_vector_abstract_api.DpfVectorAbstractAPI,
                ),
            )
            self._api.scoping_get_ids_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size
            )
            return dpf_array.DPFArray(vec) if np_array else vec.np_array.tolist()

        except NotImplementedError:
            return self._api.scoping_get_ids(self, np_array)

    def set_id(self, index, scopingid):
        """Set the ID of a scoping's index.

        Parameters
        ----------
        index : int
            Index of the scoping.
        scopingid : int
            ID of the scoping.
        """
        self._api.scoping_set_entity(self, scopingid, index)

    def _get_id(self, index):
        """Retrieve the index that the scoping ID is located on.

        Parameters
        ----------
        index : int
            Index of the scoping

        Returns
        -------
        id : int
            ID of the scoping's index.
        """
        return self._api.scoping_id_by_index(self, index)

    def _get_index(self, scopingid):
        """Retrieve an ID corresponding to an ID in the scoping.

        Parameters
        ----------
        id : int
            ID to retrieve.

        Returns
        -------
        index : int
            Index of the ID.
        """
        return self._api.scoping_index_by_id(self, scopingid)

    def id(self, index: int):
        """Retrieve the ID at a given index.

        Parameters
        ----------
        index : int
            Index for the ID.

        Returns
        -------
        size : int

        """
        return self._get_id(index)

    def index(self, id: int):
        """Retrieve the index of a given ID.

        Parameters
        ----------
        id : int
            ID for the index to retrieve.

        Returns
        -------
        size : int

        """
        return self._get_index(id)

    @property
    def ids(self):
        """Retrieve a list of IDs in the scoping.

        Returns
        -------
        ids : DPFArray, list of int
            List of IDs to retrieve. By default a mutable DPFArray is returned, to change
            the return type to a list for the complete python session, see
            :func:`ansys.dpf.core.settings.get_runtime_client_config` and
            :func:`ansys.dpf.core.runtime_config.RuntimeClientConfig.return_arrays`.
            To change the return type to a list once, use
            :func:`ansys.dpf.core.scoping.Scoping._get_ids` with the parameter ``np_array=False``.

        Notes
        -----
        Print a progress bar.
        """
        return self._get_ids()

    @ids.setter
    def ids(self, value):
        self._set_ids(value)

    @property
    def location(self):
        """Location of the IDs as a string, such as ``"nodal"``, ``"elemental"``,
        and ``"time_freq"``.

        Returns
        -------
        location : str

        """
        return self._get_location()

    @location.setter
    def location(self, value):
        self._set_location(value)

    def __len__(self):
        return self._count()

    def __del__(self):
        try:
            self._deleter_func[0](self._deleter_func[1](self))
        except Exception as e:
            print(str(e.args), str(self._deleter_func[0]))
            warnings.warn(traceback.format_exc())

    def __iter__(self):
        return self.ids.__iter__()

    def __getitem__(self, key):
        """Retrieve the ID at a requested index."""
        return self.id(key)

    def __setitem__(self, index, id):
        """Retrieve the ID at a requested index."""
        return self.set_id(index, id)

    @property
    def size(self):
        """Length of the list of IDs.

        Returns
        -------
        size : int

        """
        return self._count()

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description

        return _description(self._internal_obj, self._server)

    def deep_copy(self, server=None):
        """Create a deep copy of the scoping's data on a given server.

        This method is useful for passiong data from one server instance to another.

        Parameters
        ----------
        server : ansys.dpf.core.server, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        scoping_copy : Scoping
        """
        scop = Scoping(server=server)
        scop.ids = self.ids
        scop.location = self.location
        return scop

    def as_local_scoping(self):
        """Create a deep copy of the scoping that can be accessed and modified locally.

        This method allows you to access and modify the local copy of the scoping
        without sending a request to the server. It should be used in a ``with``
        statement so that the local field is released and the data is sent to
        the server in one action. If it is not used in a ``with`` statement,
        :func:`<release_data> Scoping.release_data()` should be used to update the scoping.

        Warning
        -------
        If this `as_local_scoping` method is not used as a context manager in a
        ``with`` statement or if the method `release_data()` is not called,
        the data will not be updated.

        Returns
        -------
        local_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> num_entities = 3
        >>> scoping_to_local = dpf.Scoping()
        >>> with scoping_to_local.as_local_scoping() as scoping:
        ...     for i in range(0,num_entities):
        ...         scoping[i] = i+1

        """  # noqa: E501
        return _LocalScoping(self)


class _LocalScoping(Scoping):
    """Caches the internal data of the scoping so that it can be modified locally.

    A single update request is sent to the server when the local scoping is deleted.

    Parameters
    ----------
    scoping : Scoping
        Scoping to copy locally.

    """

    def __init__(self, scoping):
        super(_LocalScoping, self).__init__(scoping=scoping)
        self.__cache_data__(scoping)

    def __cache_data__(self, owner_scoping):
        self._scoping_ids_copy = owner_scoping._get_ids(False)
        self._location = owner_scoping.location
        self.__init_map__()

    def __init_map__(self):
        self._mapper = dict(zip(self._scoping_ids_copy, np.arange(self._count())))

    def _count(self):
        """
        Returns
        -------
        count : int
            Number of scoping IDs.
        """
        return len(self._scoping_ids_copy)

    def _get_location(self):
        """Retrieve the location of the IDs.

        Returns
        -------
        location : str
            Location of the IDs.
        """
        return self._location

    @_setter
    def _set_location(self, loc=locations.nodal):
        """
        Parameters
        ----------
        loc : str or core.locations enum
            Location needed.

        """
        self._location = loc

    @_setter
    @version_requires("2.1")
    def _set_ids(self, ids):
        """
        Parameters
        ----------
        ids : list of int
            IDs to set.

        Notes
        -----
        Print a progress bar.
        """
        # must convert to a list for gRPC
        if isinstance(ids, range):
            ids = list(ids)
        elif isinstance(ids, (np.ndarray, np.generic)):
            ids = ids.tolist()

        self._scoping_ids_copy = ids
        self.__init_map__()

    def _get_ids(self, np_array=False):
        """
        Returns
        -------
        ids : list[int], numpy.array (if np_array==True)
            List of IDs.

        Notes
        -----
        Print a progress bar.
        """
        if np_array:
            return np.ndarray(self._scoping_ids_copy, dtype=np.int32)
        else:
            return self._scoping_ids_copy

    @_setter
    def set_id(self, index, scopingid):
        """Set the ID of a scoping's index.

        Parameters
        ----------
        index : int
            Index of the scoping.
        scopingid : int
            ID of the scoping.
        """
        init_size = self._count()
        if init_size <= index:
            for i in range(init_size, index + 1):
                self._scoping_ids_copy.append(-1)
        self._scoping_ids_copy[index] = scopingid
        self._mapper[scopingid] = index

    @_setter
    def append(self, id):
        self._scoping_ids_copy.append(id)
        self._mapper[id] = len(self) - 1

    def _get_id(self, index):
        """Retrieve the index that the scoping ID is located on.

        Parameters
        ----------
        index : int
            Index of the scoping

        Returns
        -------
        id : int
            ID of the scoping's index.
        """
        return self._scoping_ids_copy[index]

    def _get_index(self, scopingid):
        """Retrieve an ID corresponding to an ID in the scoping.

        Parameters
        ----------
        id : int
            ID to retrieve.

        Returns
        -------
        index : int
            Index of the ID.
        """
        return self._mapper[scopingid]

    def release_data(self):
        """Release the data."""
        if hasattr(self, "_is_set") and self._is_set:
            super()._set_ids(self._scoping_ids_copy)
            super()._set_location(self._location)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if tb is None:
            self._is_exited = True
            self.release_data()
        else:
            print(tb)

    def __del__(self):
        if not hasattr(self, "_is_exited") or not self._is_exited:
            self._is_exited = True
            self.release_data()
        super(_LocalScoping, self).__del__()
        pass
